from typing import *
from abc import ABC, abstractmethod
from base_aux.base_values.m3_exceptions import *
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class NestCmp_GLET_Any:
    """
    GOAL
    ----
    APPLYING COMPARISON WITH SELF INSTANCE

    BEST USAGE
    ----------
    just redefine one method __cmp__!

    WHY NOT: JUST USING ONE BY ONE EXACT METHODS?
    ---------------------------------------------
    it is more complicated then just one explicit __cmp__()!
    __cmp__ is not directly acceptable in Python! this is not a buildIn method!
    """
    __eq__ = lambda self, other: self.__cmp__(other) == 0
    # __ne__ = lambda self, other: self.__cmp__(other) != 0

    __lt__ = lambda self, other: self.__cmp__(other) < 0
    __gt__ = lambda self, other: self.__cmp__(other) > 0
    __le__ = lambda self, other: self.__cmp__(other) <= 0
    __ge__ = lambda self, other: self.__cmp__(other) >= 0

    # USING - for just raiseIf prefix!
    # FIXME: seems need to DEPRECATE? use direct EqValid_LGTE???

    # ------------------------
    check_ltgt = lambda self, other1, other2: self > other1 and self < other2
    check_ltge = lambda self, other1, other2: self > other1 and self <= other2

    check_legt = lambda self, other1, other2: self >= other1 and self < other2
    check_lege = lambda self, other1, other2: self >= other1 and self <= other2

    # ------------------------
    check_eq = lambda self, other: self == other
    check_ne = lambda self, other: self != other

    check_lt = lambda self, other: self < other
    check_le = lambda self, other: self <= other

    check_gt = lambda self, other: self > other
    check_ge = lambda self, other: self >= other

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other: Any) -> int | NoReturn:
        """
        do try to resolve Exceptions!!! sometimes it is ok to get it!!!

        RETURN
        ------
            1=self>other
            0=self==other
            -1=self<other
        """
        # NOTE: CANT APPLY ACCURACY!!!
        raise NotImplemented()

    # -----------------------------------------------------------------------------------------------------------------
    # def __eq__(self, other):
    #     return self.__cmp__(other) == 0
    #
    # def __ne__(self, other):
    #     return self.__cmp__(other) != 0
    #
    # def __lt__(self, other):
    #     return self.__cmp__(other) < 0
    #
    # def __gt__(self, other):
    #     return self.__cmp__(other) > 0
    #
    # def __le__(self, other):
    #     return self.__cmp__(other) <= 0
    #
    # def __ge__(self, other):
    #     return self.__cmp__(other) >= 0


# =====================================================================================================================
class NestCmp_GLET_DigitAccuracy:
    """
    GOAL
    ----
    apply for digital obj
    """
    # CMP_ACCURACY_VALUE_MIN: TYPING.DIGIT_FLOAT_INT_NONE = None  # TODO: add! in case of Percent!

    CMP_ACCURACY_VALUE: TYPING.DIGIT_FLOAT_INT_NONE = None
    CMP_ACCURACY_PERCENT: TYPING.DIGIT_FLOAT_INT_NONE = None

    CMP_VALUE: TYPING.DIGIT_FLOAT_INT    # property

    _cmp_accuracy__last: TYPING.DIGIT_FLOAT_INT_NONE = None

    @property
    def CMP_VALUE(self) -> TYPING.DIGIT_FLOAT_INT:
        raise NotImplementedError()

    def __init__(
            self,
            *args,
            cmp_accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            cmp_accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None,
            **kwargs,
    ) -> None | NoReturn:
        if cmp_accuracy_value is not None:
            self.CMP_ACCURACY_VALUE = cmp_accuracy_value
        if cmp_accuracy_percent is not None:
            self.CMP_ACCURACY_PERCENT = cmp_accuracy_percent

        self._cmp_accuracy__check_correctness(cmp_accuracy_value, cmp_accuracy_percent)
        self._cmp_accuracy__get_active()    # simply update _cmp_accuracy__last! no more here!
        super().__init__(*args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.CMP_VALUE}/accuracy_last={self._cmp_accuracy__last})"

    def __repr__(self) -> str:
        return str(self)

    # -----------------------------------------------------------------------------------------------------------------
    def _cmp_accuracy__check_correctness(
            self,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None,
            raised: bool = True,
    ) -> bool | NoReturn:
        """
        GOAL
        ----
        1. use only one value in one level (method/init - value)
        2. use appropriate type
        """
        raised_msg = None
        # step1 --------
        if accuracy_value is not None and accuracy_percent is not None:
            raised_msg = f"dont use both {accuracy_value=}/{accuracy_percent=}"

        elif self.CMP_ACCURACY_VALUE is not None and self.CMP_ACCURACY_PERCENT is not None:
            raised_msg = f"dont use both {self.CMP_ACCURACY_VALUE=}/{self.CMP_ACCURACY_PERCENT=}"

        # step2 --------
        for accuracy in [accuracy_value, accuracy_percent, self.CMP_ACCURACY_VALUE, self.CMP_ACCURACY_PERCENT]:
            if accuracy is not None and not isinstance(accuracy, (int, float)):
                raised_msg = f"inappropriate type {accuracy=}"
                break

        # final --------
        if raised_msg:
            print(raised_msg)
            if raised:
                raise Exc__WrongUsage(raised_msg)

        return raised_msg is None

    def _cmp_accuracy__translate_from_percent(self, accuracy_percent: float | None = None) -> float | NoReturn:
        """
        NOTE
        ----
        CAREFUL for actual values!!!
        when FLOAT+FLOAT - applied additional tailing value!!!
        dont mind it and dont use its result too strong!!!
        """
        if accuracy_percent is None:
            accuracy_percent = self.CMP_ACCURACY_PERCENT or 0

        result = self.CMP_VALUE * accuracy_percent / 100
        return result

    def _cmp_accuracy__get_active(
            self,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None
    ) -> TYPING.DIGIT_FLOAT_INT | NoReturn:     # add tests for meth
        """
        GOAL
        ----
        return final accuracy_value VALUE! from
        """
        self._cmp_accuracy__check_correctness(accuracy_value, accuracy_percent)

        result = 0
        # if accuracy_value is not None and accuracy_percent is not None:
        #     raise Exc__WrongUsage(f"dont use both {accuracy_value=}/{accuracy_percent=}")

        if accuracy_value is None and accuracy_percent is None:
            if self.CMP_ACCURACY_VALUE is not None:
                result = self.CMP_ACCURACY_VALUE
            elif self.CMP_ACCURACY_PERCENT is not None:
                result = self._cmp_accuracy__translate_from_percent(self.CMP_ACCURACY_PERCENT)
            else:
                result = 0

        elif accuracy_value is not None:
            result = accuracy_value

        elif accuracy_percent is not None:
            result = self._cmp_accuracy__translate_from_percent(accuracy_percent)

        # final check -------------------------------------
        if not isinstance(result, (int, float)):
            raise Exc__WrongUsage(f'{accuracy_value=}')

        self._cmp_accuracy__last = result
        return result

    # DEPENDANTS -------------------
    # NOTE: be careful when get Exc on second cmp with first False!
    cmp_gtlt = lambda self, other1, other2, accuracy_value=None, accuracy_percent=None: self.cmp_gt(other1, accuracy_value, accuracy_percent) and self.cmp_lt(other2, accuracy_value, accuracy_percent)
    cmp_gtle = lambda self, other1, other2, accuracy_value=None, accuracy_percent=None: self.cmp_gt(other1, accuracy_value, accuracy_percent) and self.cmp_le(other2, accuracy_value, accuracy_percent)

    cmp_gelt = lambda self, other1, other2, accuracy_value=None, accuracy_percent=None: self.cmp_ge(other1, accuracy_value, accuracy_percent) and self.cmp_lt(other2, accuracy_value, accuracy_percent)
    cmp_gele = lambda self, other1, other2, accuracy_value=None, accuracy_percent=None: self.cmp_ge(other1, accuracy_value, accuracy_percent) and self.cmp_le(other2, accuracy_value, accuracy_percent)

    cmp_eq = lambda self, other, accuracy_value=None, accuracy_percent=None: self.cmp_gele(other, other, accuracy_value, accuracy_percent)
    cmp_ne = lambda self, other, accuracy_value=None, accuracy_percent=None: not self.cmp_eq(other, accuracy_value, accuracy_percent)

    # accuracy_value DEF ------------------------
    __eq__ = lambda self, other: self.cmp_eq(other)
    # __ne__ = lambda self, other: self.__cmp__(other) != 0

    __lt__ = lambda self, other: self.cmp_lt(other)
    __gt__ = lambda self, other: self.cmp_gt(other)
    __le__ = lambda self, other: self.cmp_le(other)
    __ge__ = lambda self, other: self.cmp_ge(other)

    # BASE ------------------------------------------------------------------------------------------------------------
    def cmp_gt(
            self,
            other: TYPING.DIGIT_FLOAT_INT,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None
    ) -> bool | NoReturn:   # NoReturn is only for bad accuracy_value and insorrect (nonDigital) other!!!!
        accuracy_value = self._cmp_accuracy__get_active(accuracy_value=accuracy_value, accuracy_percent=accuracy_percent)

        result = (other - accuracy_value) < self.CMP_VALUE
        return result

    def cmp_ge(
            self,
            other: TYPING.DIGIT_FLOAT_INT,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None
    ) -> bool | NoReturn:
        accuracy_value = self._cmp_accuracy__get_active(accuracy_value=accuracy_value, accuracy_percent=accuracy_percent)

        result = (other - accuracy_value) <= self.CMP_VALUE
        return result

    def cmp_le(
            self,
            other: TYPING.DIGIT_FLOAT_INT,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None
    ) -> bool | NoReturn:
        accuracy_value = self._cmp_accuracy__get_active(accuracy_value=accuracy_value, accuracy_percent=accuracy_percent)

        result = self.CMP_VALUE <= (other + accuracy_value)
        return result

    def cmp_lt(
            self,
            other: TYPING.DIGIT_FLOAT_INT,
            accuracy_value: TYPING.DIGIT_FLOAT_INT_NONE = None,
            accuracy_percent: TYPING.DIGIT_FLOAT_INT_NONE = None
    ) -> bool | NoReturn:
        accuracy_value = self._cmp_accuracy__get_active(accuracy_value=accuracy_value, accuracy_percent=accuracy_percent)

        result = self.CMP_VALUE < (other + accuracy_value)
        return result


# =====================================================================================================================



import pytest
import requests

from base_aux.servers.m1_client_requests import Client_RequestItem, ResponseMethod, Client_RequestsStack
from base_aux.servers.m3_server2_fastapi import ServerFastApi_Thread


# =====================================================================================================================
class Test__RequestItem:
    PORT_TEST: int = 8088

    # @classmethod
    # def setup_class(cls):
    #     pass
    #
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__with_server(self):
        TEST_DATA = {'value': 1}

        # SERVER -------------------------------------
        class Server(ServerFastApi_Thread):
            PORT = self.PORT_TEST + 3

        server = Server()
        server.start()

        assert server.data.dict.get("value") is None

        # check MANUALLY ----------------------------
        response = requests.post(url=f"http://localhost:{server.PORT}/post/dict", timeout=1, json=TEST_DATA)
        assert response.json() == TEST_DATA

        # check VICTIM ------------------------------
        class VictimPost(Client_RequestItem):
            START_ON_INIT = True
            PORT = server.PORT
            ROUTE = "/post/dict"

        victim = VictimPost(body=TEST_DATA)
        victim.wait()
        assert victim.RESPONSE.ok
        assert victim.RESPONSE.json() == TEST_DATA

        class VictimGet(Client_RequestItem):
            START_ON_INIT = True
            PORT = server.PORT
            ROUTE = "/return_types/str"
            METHOD = ResponseMethod.GET

        victim = VictimGet(body=TEST_DATA)
        victim.wait()
        assert victim.RESPONSE.ok
        assert victim.RESPONSE.json() == "str"

    # -----------------------------------------------------------------------------------------------------------------
    def test__noserver(self):
        TEST_DATA = {'value': 1}
        host_wrong = "host_wrong"

        # check MANUALLY ----------------------------
        try:
            response = requests.post(url=f"http://{host_wrong}/", timeout=0.2, json=TEST_DATA)
        except:
            pass
        else:
            assert False

        # check VICTIM ------------------------------
        class VictimPost(Client_RequestItem):
            HOST = host_wrong
            START_ON_INIT = True
            TIMEOUT_SEND = 0.2

        for _ in range(2):
            victim = VictimPost(body=TEST_DATA)
            victim.wait()
            assert not victim.check_success()

    # -----------------------------------------------------------------------------------------------------------------
    # def test__noserver_timeout(self):
    #     TEST_DATA = {'value': 1}
    #     host_wrong = "host_wrong"
    #
    #     # check MANUALLY ----------------------------
    #     print()
    #     print()
    #     print()
    #     print()
    #     for timeout_item in [0.5, 1.0, 1.5]:
    #         time_start = time.time()
    #         try:
    #             response = requests.post(url=f"http://{host_wrong}/", timeout=timeout_item, json=TEST_DATA)
    #             assert False
    #         except:
    #             assert True
    #
    #         time_finish = time.time()
    #         time_period = time_finish - time_start
    #
    #         print(f"{time_start=}/{time_finish=}/{time_period=}")
    #         assert time_period >= timeout_item
    #         assert time_period < timeout_item * 3
    #
    #     # check VICTIM ------------------------------
    #     # class VictimPost(Client_RequestItem):
    #     #     HOST = host_wrong
    #     #     START_ON_INIT = True
    #     #     TIMEOUT_SEND = 0.5
    #     #
    #     #
    #     # for _ in range(1):
    #     #     victim = VictimPost(body=TEST_DATA)
    #     #     victim.wait()
    #     #     assert not victim.check_success()


# =====================================================================================================================
class Test__RequestsStack:
    PORT_TEST: int = 8088

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__1(self):
        TEST_DATA = {'value': 1}

        # SERVER -------------------------------------
        class Server(ServerFastApi_Thread):
            PORT = self.PORT_TEST + 4

        server = Server()
        server.start()

        # assert server.data.dict.get("value") is None

        # check MANUALLY ----------------------------
        response = requests.post(url=f"http://localhost:{server.PORT}/post/dict", timeout=1, json=TEST_DATA)
        assert response.json() == TEST_DATA

        # check VICTIM ------------------------------
        class ClientRequestItem_1(Client_RequestItem):
            PORT = server.PORT
            ROUTE = "/post/dict"

        class Victim(Client_RequestsStack):
            REQUEST_CLS = ClientRequestItem_1

        victim = Victim()

        assert server.data.dict.get("value") == 1
        victim.send(body={'value': 2})
        victim.wait()
        assert server.data.dict.get("value") == 2

    # -----------------------------------------------------------------------------------------------------------------
    def test__2_noserver(self):
        SEND_COUNT = 100
        TEST_DATA = {'value': 1}
        host_wrong = "host_wrong"

        # check MANUALLY ----------------------------
        try:
            response = requests.post(url=f"http://{host_wrong}/", timeout=0.2, json=TEST_DATA)
        except:
            pass
        else:
            assert False

        # check VICTIM ------------------------------
        class ClientRequestItem_1(Client_RequestItem):
            HOST = host_wrong
            TIMEOUT_SEND = 0.2

        class Victim(Client_RequestsStack):
            REQUEST_CLS = ClientRequestItem_1

        victim = Victim()

        for index in range(SEND_COUNT):
            victim.send(body={'index': index})
            assert not victim.check_success()

        victim.wait()
        assert not victim.check_success()
        assert len(victim.stack) == SEND_COUNT


# =====================================================================================================================



from base_aux.breeders.m1_breeder_str1_series import *


# =====================================================================================================================
class Test__BreederStrSeries:
    @classmethod
    def setup_class(cls):
        cls.Victim = BreederStrSeries

    # # @classmethod
    # # def teardown_class(cls):
    # #     pass
    # #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__START_OUTER_None(self):
        victim = self.Victim(None, 1, "test_%s", 1)
        assert victim.START_OUTER == None
        assert victim.COUNT == 1
        assert victim.TEMPLATE == "test_%s"
        assert victim.START_INNER == 1

        try:
            assert victim.get_dict__inner()
            assert False
        except Exc__StartOuterNONE_UsedInStackByRecreation:
            assert True

        try:
            assert victim.get_dict__outer()
            assert False
        except Exc__StartOuterNONE_UsedInStackByRecreation:
            assert True

    def test__init_params(self):
        victim = self.Victim(0, 1, "test_%s", 1)
        assert victim.START_OUTER == 0
        assert victim.COUNT == 1
        assert victim.TEMPLATE == "test_%s"
        assert victim.START_INNER == 1

        assert victim.get_dict__inner() == {1: "test_1"}
        assert victim.get_dict__outer() == {0: "test_1"}

        victim = self.Victim(1, 1, "test_%s", 1)
        assert victim.START_OUTER == 1
        assert victim.COUNT == 1
        assert victim.TEMPLATE == "test_%s"
        assert victim.START_INNER == 1

        victim = self.Victim(1, 2, "test_%s", 1)
        assert victim.START_OUTER == 1
        assert victim.COUNT == 2
        assert victim.TEMPLATE == "test_%s"
        assert victim.START_INNER == 1

    def test__dicts(self):
        victim = self.Victim(0, 1, "test_%s", 0)
        assert victim.get_dict__inner() == {0: "test_0"}
        assert victim.get_dict__outer() == {0: "test_0"}

        victim = self.Victim(0, 1, "test_%s", 1)
        assert victim.get_dict__inner() == {1: "test_1"}
        assert victim.get_dict__outer() == {0: "test_1"}

        victim = self.Victim(0, 2, "test_%s", 1)
        assert victim.get_dict__inner() == {1: "test_1", 2: "test_2"}
        assert victim.get_dict__outer() == {0: "test_1", 1: "test_2"}

    def test__contains(self):
        # --------------------------------------------------
        victim = self.Victim(0, 1, "test_%s", 1)

        # indexes
        assert 0 in victim
        assert 1 not in victim
        assert 2 not in victim

        # values
        assert "test_0" not in victim
        assert "test_1" in victim
        assert "test_2" not in victim

        # --------------------------------------------------
        victim = self.Victim(0, 2, "test_%s", 1)

        # indexes
        assert 0 in victim
        assert 1 in victim
        assert 2 not in victim

        # values
        assert "test_0" not in victim
        assert "test_1" in victim
        assert "test_2" in victim

        # --------------------------------------------------
        victim = self.Victim(1, 2, "test_%s", 0)

        # indexes
        assert 0 not in victim
        assert 1 in victim
        assert 2 in victim

        # values
        assert "test_0" in victim
        assert "test_1" in victim
        assert "test_2" not in victim

    def test__getitem(self):
        # --------------------------------------------------
        victim = self.Victim(0, 1, "test_%s", 0)
        # indexes
        assert victim[0] == "test_0"
        try:
            result = victim[1]
        except:
            pass
        else:
            assert False

        # values
        assert victim["test_0"] == 0
        try:
            result = victim["test_1"]
        except:
            pass
        else:
            assert False

        # --------------------------------------------------
        victim = self.Victim(0, 1, "test_%s", 1)
        # indexes
        try:
            result = victim[0]
        except:
            pass
        else:
            assert False
        assert victim[1] == "test_1"


        # values
        try:
            result = victim["test_0"]
        except:
            pass
        else:
            assert False
        assert victim["test_1"] == 1

    def test__listed_index__by_outer(self):
        # --------------------------------------------------
        victim = self.Victim(0, 2, "test_%s", 0)
        assert victim.get_dict__inner() == {0: "test_0", 1: "test_1"}
        assert victim.get_dict__outer() == {0: "test_0", 1: "test_1"}

        assert list(victim.get_dict__outer()) == [0, 1]
        assert list(victim.get_dict__inner()) == [0, 1]

        assert victim.get_listed_index__by_outer(0) == 0
        assert victim.get_listed_index__by_outer(1) == 1
        try:
            assert victim.get_listed_index__by_outer(2)
        except:
            pass
        else:
            assert False

        # --------------------------------------------------
        victim = self.Victim(0, 2, "test_%s", 1)
        assert victim.get_dict__inner() == {1: "test_1", 2: "test_2"}
        assert victim.get_dict__outer() == {0: "test_1", 1: "test_2"}

        assert list(victim.get_dict__outer()) == [0, 1]
        assert list(victim.get_dict__inner()) == [1, 2]

        assert victim.get_listed_index__by_outer(0) == 0
        assert victim.get_listed_index__by_outer(1) == 1
        try:
            assert victim.get_listed_index__by_outer(2)
        except:
            pass
        else:
            assert False

    def test__listed_index__by_value(self):
        # --------------------------------------------------
        victim = self.Victim(0, 2, "test_%s", 0)
        assert victim.get_dict__inner() == {0: "test_0", 1: "test_1"}
        assert victim.get_dict__outer() == {0: "test_0", 1: "test_1"}

        assert victim.get_listed_index__by_value("test_0") == 0
        assert victim.get_listed_index__by_value("test_1") == 1
        try:
            assert victim.get_listed_index__by_value("test_2")
        except:
            pass
        else:
            assert False

        # --------------------------------------------------
        victim = self.Victim(0, 2, "test_%s", 1)
        assert victim.get_dict__inner() == {1: "test_1", 2: "test_2"}
        assert victim.get_dict__outer() == {0: "test_1", 1: "test_2"}

        try:
            assert victim.get_listed_index__by_value("test_0")
        except:
            pass
        else:
            assert False
        assert victim.get_listed_index__by_value("test_1") == 0
        assert victim.get_listed_index__by_value("test_2") == 1


# =====================================================================================================================



from base_aux.aux_eq.m2_eq_aux import *
from base_aux.aux_text.m1_text_aux import *
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class Validators:
    """
    CAREFULL
    --------
    using one validators inside others - is not so simple!

    GOAL
    ----
    collect all validators (funcs) in one place
    applicable in Base_EqValid only (by common way), but you can try using it separated!

    SPECIALLY CREATED FOR
    ---------------------
    Base_EqValid

    RULES
    -----
    1/ NoReturn - available for all returns as common!!! but sometimes it cant be reached (like TRUE/RAISE)
    2/ other_final - always at first place! other params goes nest (usually uncovered)
    3/ in cmp result KEEP variant ALWAYS AT FIRST PLACE!!! to use its DANDER-methods first! like in CMP_LT

    """
    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def IsinstanceSameinstance(self, other_final: Any, variant: type[Any] | Any) -> bool | NoReturn:
        """
        GOAL
        ----
        isinstance or SameInstance!!!
        """
        try:
            issubclass(variant, object)
        except:
            variant = variant.__class__

        return isinstance(other_final, variant)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def Contain(self, other_final: Any, variant: Any) -> bool | NoReturn:
        """
        GOAL
        ----
        check each variant with other by IN operator
        mainly using for check substrs (variants) in BaseStr

        SPECIALLY CREATED FOR
        ---------------------
        AttrsAux.dump_dict/AttrsDump to skip exact attrs with Parts in names
        """
        return variant in other_final

    @staticmethod
    def ContainStrIc(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return str(variant).lower() in str(other_final).lower()

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def Startswith(self, other_final: Any, variant: Any) -> bool | NoReturn:
        other_final = str(other_final)
        variant = str(variant)
        return other_final.startswith(variant)

    @staticmethod
    def StartswithIc(self, other_final: Any, variant: Any) -> bool | NoReturn:
        other_final = str(other_final).lower()
        variant = str(variant).lower()
        return other_final.startswith(variant)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def Endswith(self, other_final: Any, variant: Any) -> bool | NoReturn:
        other_final = str(other_final)
        variant = str(variant)
        return other_final.endswith(variant)

    @staticmethod
    def EndswithIc(self, other_final: Any, variant: Any) -> bool | NoReturn:
        other_final = str(other_final).lower()
        variant = str(variant).lower()
        return other_final.endswith(variant)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def BoolTrue(self, other_final: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object called with no raise and no Exception in result
        """
        if self.OTHER_RAISED or TypeAux(other_final).check__exception():
            return False
        try:
            return bool(other_final)
        except:
            return False

    # TODO: add FALSE????? what to do with exc and real false?

    @staticmethod
    def Raise(self, other_final: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object called with raised
        if other is exact final Exception without raising - it would return False!
        """
        return self.OTHER_RAISED

    @staticmethod
    def NotRaise(self, other_final: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object called with raised
        if other is exact final Exception without raising - it would return False!
        """
        return not self.OTHER_RAISED

    @staticmethod
    def Exc(self, other_final: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object is exact Exception or Exception()
        if raised - return False!!
        """
        return not self.OTHER_RAISED and TypeAux(other_final).check__exception()

    @staticmethod
    def ExcRaise(self, other_final: Any) -> bool:
        """
        GOAL
        ----
        True - if Other object is exact Exception or Exception() or Raised
        """
        return self.OTHER_RAISED or TypeAux(other_final).check__exception()

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def CMP_EQ(self, other_final: Any, variant: Any) -> bool | NoReturn:
        print(f"CMP_EQ={other_final=}/{variant=}")
        return variant == other_final

    @staticmethod
    def CMP_EQ__StrIc(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return str(variant).lower() == str(other_final).lower()

    @staticmethod
    def CMP_EQ__NumParsedSingle(self, other_final: Any, variant: Any) -> bool | NoReturn:
        other_final = TextAux(other_final).parse__number_single()
        print(f"CMP_EQ__NumParsedSingle={other_final=}/{variant=}")
        return Validators.CMP_EQ(self, other_final, variant)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def CMP_LT(self, other_final: Any, variant: Any, parse__number_single: bool = None) -> bool | NoReturn:
        if parse__number_single:
            other_final = TextAux(other_final).parse__number_single()
        return variant > other_final    # NOTE: KEEP variant ALWAYS AT FIRST PLACE!!! to use its DANDER-methods!

    @staticmethod
    def CMP_LE(self, other_final: Any, variant: Any, parse__number_single: bool = None) -> bool | NoReturn:
        if parse__number_single:
            other_final = TextAux(other_final).parse__number_single()
        return variant >= other_final

    @staticmethod
    def CMP_GT(self, other_final: Any, variant: Any, parse__number_single: bool = None) -> bool | NoReturn:
        if parse__number_single:
            other_final = TextAux(other_final).parse__number_single()
        return variant < other_final

    @staticmethod
    def CMP_GE(self, other_final: Any, variant: Any, parse__number_single: bool = None) -> bool | NoReturn:
        if parse__number_single:
            other_final = TextAux(other_final).parse__number_single()
        return variant <= other_final

    @staticmethod
    def CMP_LGTE(
            self,
            other_final: Any,
            lt: Any | None = None,
            le: Any | None = None,
            gt: Any | None = None,
            ge: Any | None = None,
            parse__number_single: bool = None,
    ) -> bool | NoReturn:
        """
        NOTE
        ----
        used all variants of cmp (l/g* + *t/e) just to make a one clear logical func!
        intended using only one or two appropriate combination!
        """
        if parse__number_single:
            other_final = TextAux(other_final).parse__number_single()

        for validator, variant in [
            (Validators.CMP_LT, lt),
            (Validators.CMP_LE, le),
            (Validators.CMP_GT, gt),
            (Validators.CMP_GE, ge),
        ]:
            if variant is not None:
                if not validator(self, other_final, variant):
                    return False
        return True

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def CMP_LT_NumParsedSingle(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return Validators.CMP_LT(self, other_final, variant, parse__number_single=True)

    @staticmethod
    def CMP_LE_NumParsedSingle(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return Validators.CMP_LE(self, other_final, variant, parse__number_single=True)

    @staticmethod
    def CMP_GT_NumParsedSingle(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return Validators.CMP_GT(self, other_final, variant, parse__number_single=True)

    @staticmethod
    def CMP_GE_NumParsedSingle(self, other_final: Any, variant: Any) -> bool | NoReturn:
        return Validators.CMP_GE(self, other_final, variant, parse__number_single=True)

    @staticmethod
    def CMP_LGTE_NumParsedSingle(
            self,
            other_final: Any,
            lt: Any | None = None,
            le: Any | None = None,
            gt: Any | None = None,
            ge: Any | None = None,
    ) -> bool | NoReturn:
        return Validators.CMP_LGTE(self, other_final, lt=lt, le=le, gt=gt, ge=ge, parse__number_single=True)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def NumParsedSingle_Success(self, other_final) -> bool:
        other_final = TextAux(other_final).parse__number_single()
        return other_final is not None

    @staticmethod
    def NumParsedSingle_TypeInt(self, other_final) -> bool:
        other_final = TextAux(other_final).parse__number_single()
        return isinstance(other_final, int)

    @staticmethod
    def NumParsedSingle_TypeFloat(self, other_final) -> bool:
        other_final = TextAux(other_final).parse__number_single()
        return isinstance(other_final, float)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def Regexp(
            self,
            other_final,
            pattern: str,
            ignorecase: bool = True,
            match_link: Callable = re.fullmatch,
    ) -> bool | NoReturn:
        # NOTE: just a link!
        #   you can use directly match_link in Base_EqValid!!!!??? - no you need use it at least over LAMBDA!
        result = match_link(pattern=str(pattern), string=str(other_final), flags=re.RegexFlag.IGNORECASE if ignorecase else 0)
        return result is not None

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def AttrsByKwargs(
            self,
            other_final,
            # callable_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
            **kwargs: TYPING.KWARGS_FINAL
    ) -> bool | NoReturn:
        for key, value in kwargs.items():
            value_expected = Lambda(value).resolve__style(EnumAdj_CallResolveStyle.EXC)
            value_other = AttrAux_Existed(other_final).gai_ic__callable_resolve(key, EnumAdj_CallResolveStyle.EXC)
            if not EqAux(value_expected).check_doubleside__bool(value_other):
                return False

        # FINISH -----
        return True

    @staticmethod
    def AttrsByObj(
            self,
            other_final,
            # callable_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
            source: Any,
            # attr_level: EnumAdj_AttrScope = EnumAdj_AttrScope.NOT_PRIVATE,
    ) -> bool | NoReturn:
        for key in AttrAux_Existed(source).iter__names_filter(self.ATTR_LEVEL):
            value_expected = AttrAux_Existed(source).gai_ic__callable_resolve(key, EnumAdj_CallResolveStyle.EXC)
            value_other = AttrAux_Existed(other_final).gai_ic__callable_resolve(key, EnumAdj_CallResolveStyle.EXC)
            if not EqAux(value_expected).check_doubleside__bool(value_other):
                return False

        # FINISH -----
        return True

    # NOTE: INAPPROPRIATE!!!!
    # def AttrsByObjNotPrivate(
    #         self,
    #         other_final,
    #         # callable_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
    #         source: Any,
    # ) -> bool | NoReturn:
    #     return self._AttrsByObj(other_final=other_final, source=source, attr_level=EnumAdj_AttrScope.NOT_PRIVATE)
    # def AttrsByObjNotHidden(
    #         self,
    #         other_final,
    #         # callable_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
    #         source: Any,
    # ) -> bool | NoReturn:
    #     return self._AttrsByObj(other_final=other_final, source=source, attr_level=EnumAdj_AttrScope.NOT_HIDDEN)

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def AnnotsAllExists(
            self,
            other_final,
    ) -> bool | NoReturn:
        return AttrAux_AnnotsAll(other_final).annots__check_all_defined()


# =====================================================================================================================



import requests
from aiohttp import web

from base_aux.servers.m2_server1_aiohttp import ServerAiohttpBase
from base_aux.aux_argskwargs.m4_kwargs_eq_expect import *


# =====================================================================================================================
@pytest.mark.skip   # FIXME: BROKEN!
@pytest.mark.skipif(KwargsEqExpect_OS().bool_if__any_true(linux=True), reason="WindowsOnly if start in no main thread")
class Test__ServerAiohttp:
    # FIXME: CANT SEPARATE TESTS!!!! WORK ONLY one BY one
    PORT_TEST: int = 8081

    # @classmethod
    # def setup_class(cls):
    #     pass
    #
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__ALL(self):
        # PREPARE ==============================================
        TEST_DATA = {'test_data_key': 1}

        class Victim(ServerAiohttpBase):
            PORT = self.PORT_TEST
            test_data = {}

            async def response_post__test_post(self, request) -> web.Response:
                self.test_data = await request.json()
                return web.json_response(data=self.test_data)

            async def response_get_json__test_get_json(self, request) -> web.Response:
                return web.json_response(data=self.test_data)

        self.victim = Victim()

        class Victim2(ServerAiohttpBase):
            PORT = self.PORT_TEST + 2

        self.victim_2 = Victim2()

        # WORK ==============================================
        self.victim.start()
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/", timeout=1)

        # double start DEPandant -------------------------
        try:
            self.victim.run()
        except:
            pass
        else:
            assert False

        # double start INDEPandant ------------------------
        self.victim_2.start()
        response = requests.get(url=f"http://localhost:{self.victim_2.PORT}/", timeout=1)

        # test__GET_INDEX ----------------------------------
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/api_index", timeout=1)
        assert f'/api_index' in self.victim._ROUTES["get_html"]
        assert f'/api_index' in response.text

        # test__POST ----------------------------------
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/test_get_json", timeout=1)
        assert response.json() == {}

        response = requests.post(url=f"http://localhost:{self.victim.PORT}/test_post", timeout=1, json=TEST_DATA)
        assert f'/test_post' in self.victim._ROUTES["post"]
        assert self.victim.test_data == TEST_DATA
        assert response.json() == TEST_DATA

        # test__GET_JSON ----------------------------------
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/test_get_json", timeout=1)
        assert f'/test_get_json' in self.victim._ROUTES["get_json"]
        assert f'/{self.victim._ROUTE_NAME_PREFIX_HTML_FOR_JSON}test_get_json' in self.victim._ROUTES["get_html"]
        assert response.json() == TEST_DATA

        response = requests.get(url=f"http://localhost:{self.victim.PORT}/{self.victim._ROUTE_NAME_PREFIX_HTML_FOR_JSON}test_get_json", timeout=1)
        assert 'test_data_key' in response.text

        # test__404 ----------------------------------
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/test404", timeout=1)
        assert not response.ok
        assert response.status_code == 404

        # test__STOP/START ----------------------------------
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/", timeout=0.3)
        assert response.ok
        assert response.status_code == 200

        self.victim._app.shutdown()
        try:
            response = requests.get(url=f"http://localhost:{self.victim.PORT}/", timeout=0.3)
        except:
            pass
        else:
            assert False

        self.victim._app.startup()
        response = requests.get(url=f"http://localhost:{self.victim.PORT}/", timeout=0.3)
        assert response.ok
        assert response.status_code == 200


# =====================================================================================================================



from PyQt5.QtCore import *
from PyQt5.QtGui import *

from base_aux.breeders.m1_breeder_str2_stack import *
from base_aux.breeders.m1_breeder_str1_series import *
from .m0_base import Data_
from base_aux.translator.m1_translator import Translator


# =====================================================================================================================
class Headers(BreederStrStack):
    NAME: int = 0
    BATCH: BreederStrSeries = BreederStrSeries(1, 4)


# =====================================================================================================================
class Base_TableModel(QAbstractTableModel):
    DATA: Data_
    HEADERS: BreederStrStack = Headers()
    HTRANSLATOR: Translator = None

    # METHODS USER ----------------------------------------------------------------------------------------------------
    def __init__(self, data: Optional[Any] = None):
        super().__init__(parent=None)
        self.DATA = data

    def _data_reread(self) -> None:
        """
        just redraw model by reread all data!
        """
        self.endResetModel()

    # =================================================================================================================
    def rowCount(self, parent: Any = None, *args, **kwargs) -> int:
        return len(self.DATA.ROWS)  # + 1  # [+1]for finalResults

    def columnCount(self, parent: Any = None, *args, **kwargs) -> int:
        return len(self.DATA.DEVS) + 1

    # =================================================================================================================
    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> Any:
        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DisplayRole:      # in headerData WORK ONLY DisplayRole+TOOLTIP!!!
            if orientation == Qt.Horizontal:
                value = self.HEADERS[section]
                if self.HTRANSLATOR:
                    value = self.HTRANSLATOR(value)
                return value

            if orientation == Qt.Vertical:
                return str(section + 1)

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.ToolTipRole:
            if orientation == Qt.Horizontal:
                return self.HEADERS[section]

        # # -------------------------------------------------------------------------------------------------------------
        # if role == Qt.CheckStateRole:      # ЧЕКБОКСЫ  - IS IT WORK???
        #     # -------------------
        #     dut = None
        #     if section > 0:
        #         return Qt.Checked
        #
        #         dut = self.DATA.DEVS[section - 1]
        #
        #         if section % 2:
        #             return Qt.Unchecked
        #         else:
        #             return Qt.Checked

        # -------------------------------------------------------------------------------------------------------------
        # if role == Qt.BackgroundColorRole:  # DONT WORK!!!
        #     return QColor('red')
        #
        #     # -------------------
        #     dut = None
        #     if section > 1:
        #         dut = self.DATA.DEVS[section - 1]
        #         if not hasattr(dut, "SKIP"):
        #             dut.SKIP = False
        #
        #         if not dut.SKIP:
        #             return QColor('#f2f2f2')

    # def setHeaderData(self, section: int, orientation: Qt.Orientation, value: Any, role: int = Qt.DisplayRole) -> bool:
    #     # -------------------------------------------------------------------------------------------------------------
    #     if orientation == Qt.Vertical:
    #         return super().setHeaderData(section, orientation, value, role)
    #
    #     # -------------------------------------------------------------------------------------------------------------
    #     if orientation == Qt.Horizontal:
    #         # -------------------
    #         dut = None
    #         if section > 0:
    #             dut = self.DATA.DEVS[section - 1]
    #
    #         # -------------------
    #         if role == Qt.CheckStateRole:      # ЧЕКБОКСЫ
    #             if section == 1:
    #                 return "NAME"
    #             if section > 0:
    #                 return f"{section}"
    #
    #     return True

    # =================================================================================================================
    def flags(self, index: QModelIndex) -> int:
        """
        VARIANTS FLAGS
        --------------
        flags = Qt.NoItemFlags              # 0=без флагов - полное отключение и деактивация всего
        flags |= Qt.ItemIsSelectable        # 1=выделяется цветом при выборе, иначе только внешней рамкой!
        flags |= Qt.ItemIsEditable          # 2=можно набирать с клавиатуры!
        flags |= Qt.ItemIsDragEnabled       # 4=
        flags |= Qt.ItemIsDropEnabled       # 8=
        flags |= Qt.ItemIsUserCheckable     # 16=для чекбоксов дает возможность их изменять мышью!
        flags |= Qt.ItemIsEnabled           # 32=если нет - будет затенен! без возможности выбора!
        flags |= Qt.ItemIsTristate=ItemIsAutoTristate    # 64=включение промежуточного значения чекбокса
        flags |= Qt.ItemNeverHasChildren    # 128=
        flags |= Qt.ItemIsUserTristate      # 256=
        """
        col = index.column()
        row = index.row()

        flags = super().flags(index)    # recommended using as default! and switching exact flags

        if col == 0:
            # FIXME: HEADER CAUSE EXCEPTION CLOSE!

            flags = Qt.NoItemFlags  # 0=без флагов - полное отключение и деактивация всего
            # flags |= Qt.ItemIsSelectable  # 1=выделяется цветом при выборе, иначе только внешней рамкой!
            # flags |= Qt.ItemIsEditable  # 2=можно набирать с клавиатуры!
            # flags |= Qt.ItemIsDragEnabled  # 4=
            # flags |= Qt.ItemIsDropEnabled  # 8=
            flags |= Qt.ItemIsUserCheckable  # 16=для чекбоксов дает возможность их изменять мышью!
            flags |= Qt.ItemIsEnabled  # 32=если нет - будет затенен! без возможности выбора!
            flags |= Qt.ItemIsTristate  #=ItemIsAutoTristate  # 64=включение промежуточного значения чекбокса
            # flags |= Qt.ItemNeverHasChildren  # 128=
            # flags |= Qt.ItemIsUserTristate  # 256=

        return flags

    # =================================================================================================================
    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        # if not index.isValid():
        #     return QVariant()
        """
        ROLES
        -------------
        DisplayRole = 0
        DecorationRole = 1
        EditRole = 2
        ToolTipRole = 3
        StatusTipRole = 4
        WhatsThisRole = 5
        FontRole = 6
        TextAlignmentRole = 7

        BackgroundRole=BackgroundColorRole = 8
        ForegroundRole=TextColorRole = 9

        CheckStateRole = 10
        SizeHintRole = 13
        InitialSortOrderRole = 14
        UserRole = 256
        """
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        tc = list(self.DATA.ROWS)[row]
        dut = None
        if col in self.HEADERS.BATCH:
            dut = self.DATA.DEVS[col-1]

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DisplayRole:
            if col < 0:
                print(f"STOP{index=}")
            if col == 0:
                return f'{tc.NAME}'
            if col in self.HEADERS.BATCH:
                return f'{dut.result}'

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextAlignmentRole:
            """
            VARIANTS ALIGN
            --------------
            not exists NAME!!!} = 0         # (LEFT+TOP) [[[[[[[[DEFAULT IS [LEFT+TOP]]]]]]]]]

            AlignLeft=AlignLeading = 1      # LEFT(+TOP)
            AlignRight=AlignTrailing = 2    # RIGHT(+TOP)

            AlignTop = 32       # TOP(+LEFT)
            AlignBottom = 64    # BOT(+LEFT)

            AlignHCenter = 4    # HCENTER(+TOP)
            AlignVCenter = 128  # VCENTER(+LEFT)
            AlignCenter = 132   # VCENTER+HCENTER

            # =====MAYBE DID NOT FIGURED OUT!!!
            AlignAbsolute = 16      # (LEFT+TOP) == asDEFAULT
            AlignBaseline = 256     # (LEFT+TOP) == asDEFAULT

            AlignJustify = 8        # (LEFT+TOP) == asDEFAULT

            AlignHorizontal_Mask = 31   # TOP+RIGHT
            AlignVertical_Mask = 480    # LEFT+VCENTER
            """
            if col == 0:
                return Qt.AlignVCenter
            if col in self.HEADERS.BATCH:
                return Qt.AlignCenter

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.FontRole:
            if row % 2:
                # QFont("Arial", 9, QFont.Bold)
                font = QFont()

                font.setBold(True)
                font.setItalic(True)

                font.setOverline(True)      # надчеркнутый
                font.setStrikeOut(True)     # зачеркнутый
                font.setUnderline(True)     # подчеркнутый

                # не понял!! --------------------
                # font.setStretch(5)
                # font.setCapitalization()

                return font

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.BackgroundColorRole:   # =BackgroundRole
            if tc.SKIP or (dut and dut.SKIP):
                return QColor('#e2e2e2')

            if col in self.HEADERS.BATCH:
                if tc.result is True:
                    return QColor("green")
                if tc.result is False:
                    return QColor("red")

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextColorRole:   # =ForegroundRole
            if tc.SKIP:
                return QColor('#a2a2a2')

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:   # для чекбоксов!
            """
            VARIANTS CHECK
            --------------
            Unchecked (или 0) - флажок сброшен;
            PartiallyChecked (или 1) - флажок частично установлен;
            Checked (или 2) - флажок установлен
            """
            if col == 0:
                if tc.SKIP is True:
                    return Qt.Unchecked
                elif tc.SKIP is False:
                    return Qt.Checked
                else:
                    return Qt.PartiallyChecked

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.ToolTipRole:
            return f"{row}/{col}"

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DecorationRole:   # ИКОНКА в ячейке слева!
            # может не работать! не работало с первого раза, НО потом вдруг заработало само собой когда добавил SizeHintRole!!! больше не пропадало!!!
            if col == 1 and not tc.SKIP:
                icon = QIcon()
                icon.addPixmap(QPixmap('logo.jpg'))
                return icon

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.SizeHintRole:     # размер ячейки!
            """
            IMPORTANT: 
            1. meaning like MinimumSize!
            2. BUT if there are IconOrCheck exists - size will be same as before adding it and HIDING oversized text
            """
            if col in self.HEADERS.BATCH:
                return QSize(5, 5)

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.EditRole) -> bool:
        """
        NOTE: при фактическом изменении ОБЯЗАТЕЛЬНО возвращать TRUE!!! Иначе Exc!!!!
        in other cases you can return True either, or None!
        """
        # TODO: ALWAYS START data_reread after any SETDATA
        # if not index.isValid():
        #     return False

        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        tc = list(self.DATA.ROWS)[row]
        if col in self.HEADERS.BATCH:
            dut = self.DATA.DEVS[col-1]
        else:
            dut = None

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:      # ЧЕКБОКСЫ
            # need used flag ItemIsUserCheckable!
            if col == 0:
                tc.SKIP = value == Qt.Unchecked

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.EditRole:
            # need used flag ItemIsEditable!
            if col == 0:
                print("EditRole")

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.SizeHintRole:
            # хотел увидеть изменение размера НО ЭТО ТАК НЕ РАБОТАЕТ!!!
            print("SizeHintRole")

        # FINAL -------------------------------------------------------------------------------------------------------
        self._data_reread()
        return True


# =====================================================================================================================



from base_aux.aux_iter.m1_iter_aux import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
class Meta_ClsGaAnnotNamesAsValuesIc(type):
    """
    return from class just name of annotation as string value.
    if no corresponding annotation - raise!
    """
    _IC = EnumAdj_IgnoreCase.IGNORECASE    # NOTE: DONT USE ANNOTATIONS here!!!! _IC: EnumAdj_IgnoreCase

    # IC DEPENDENT ---------------------------------------
    def __getattr__(cls, item: str) -> str | NoReturn:
        return cls[item]

    def __getitem__(cls, item: int | str | Any) -> str | NoReturn:
        annots = AttrAux_AnnotsAll(cls).dump_dict__annot_types()
        try:
            item = int(item)
        except:
            item = str(item)

        if isinstance(item, int):
            return list(annots)[item]
        else:
            if cls._IC == EnumAdj_IgnoreCase.IGNORECASE:
                return IterAux(annots).item__get_original(item, _raise=True)
            if item in annots:
                return item
            else:
                msg = f"[ERROR] META:'{cls.__name__}' CLASS has no annotation for '{item=}'"
                raise AttributeError(msg)

    def __contains__(cls, item: str) -> bool:
        annots = AttrAux_AnnotsAll(cls).dump_dict__annot_types()
        if cls._IC == EnumAdj_IgnoreCase.IGNORECASE:
            return IterAux(annots).item__check_exist(item)
        else:
            return item in annots

    # IC independent ---------------------------------------
    def __iter__(cls) -> Iterable[str]:
        annots = AttrAux_AnnotsAll(cls).dump_dict__annot_types()
        yield from annots

    def __len__(cls) -> int:
        annots = AttrAux_AnnotsAll(cls).dump_dict__annot_types()
        return len(annots)

    def __str__(cls) -> str:
        annots = AttrAux_AnnotsAll(cls).dump_dict__annot_types()
        return str(tuple(annots))

    def __repr__(cls) -> str:
        return f"{cls.__name__}{cls}"


class Meta_ClsGaAnnotNamesAsValuesCs(Meta_ClsGaAnnotNamesAsValuesIc):
    """
    return from class just name of annotation as string value.
    if no corresponding annotation - raise!
    """
    _IC = EnumAdj_IgnoreCase.CASESENSE


# ---------------------------------------------------------------------------------------------------------------------
class NestGaCls_AnnotNamesAsValuesIc(metaclass=Meta_ClsGaAnnotNamesAsValuesIc):
    """
    used as simple string data container, same as OneWordStringsList with access to values by dot.
    ATTEMPT to get rid of bare data like list[str] ot tuple[str]!

    BEST EXAMPLE (Designed specially for this case)
    -----------------------------------------------
    Suppose you have
        States = ("ON", "OFF")
    next you think it would be better using it over DOT access, so you create class
        class States:
            ON = "ON"
            OFF = "OFF"

    just replace it for
        class States(NestGaCls_AnnotNamesAsValuesIc):
            ON: str
            OFF: str

    ADVANTAGES
    ----------
    1. DRY

    USAGE
    -----
    1. USE ANNOTATED ATTRIBUTES!
        class MyValues(NestGaCls_AnnotNamesAsValuesIc):
            VALUE1: str
            VALUE2: str

        print(MyValues.value1)  # AttributeError(...)
        print(MyValues.VALUE1)  # "VALUE1"
        print(MyValues.VALUE2)  # "VALUE2"
        print(MyValues.VALUE3)  # AttributeError(...)

    2. DONT SET VALUES! it would break the main idea! (but maybe you want it).
        class MyValues(NestGaCls_AnnotNamesAsValuesIc):
            VALUE1: str
            VALUE2: str = 123

        print(MyValues.VALUE1)  # "VALUE1"
        print(MyValues.VALUE2)  # 123

    3. ANNOTATING TYPE IS NOT IMPORTANT! cause of no values exists!
        class MyValues(NestGaCls_AnnotNamesAsValuesIc):
            VALUE1: Any
            VALUE2: Any

        print(MyValues.VALUE1)  # "VALUE1"
        print(MyValues.VALUE2)  # "VALUE2"

    4. CANT USE INSTANCES!
        class MyValues(NestGaCls_AnnotNamesAsValuesIc):
            VALUE1: str
            VALUE2: str

        print(MyValues.VALUE1)      # "VALUE1"
        print(MyValues().VALUE1)    #AttributeError: 'MyValues' object has no attribute 'VALUE1'

    WHY NOT ENUM?
    -------------
    0. enum is the next level but we need previous one.
    1. sometimes Enum is OverComplicated just for simple string values.
    2. usually enum use int value for name and you can change value and dont understand what it was before according for value=1
    3. in enum you will get key/value name by unnecessary attributes like *.name/value
    4. if we need only value and value is already equal to name so why we need value with name?
    5. enum keep instances, not the final results.
    6. in enum values are abstract (not valuable)! so why we need to use them if we dont care? and why we need to create not used extra data?
    FINALLY: so here you can directly accessing to exact string values by class attributes.

    WHY NOT NAMEDTUPLE?
    -------------------
    1. its again!.. why we need duplicating values if we already have it in names?
    all we want - access by '*.attribute' to the value,
    so minimum we need is using annotations, that's enough for IDE checker and return string values!
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
class NestGaCls_AnnotNamesAsValuesCs(metaclass=Meta_ClsGaAnnotNamesAsValuesCs):
    pass


# =====================================================================================================================



from base_aux.base_values.m4_primitives import *

from base_aux.aux_argskwargs.m3_args_bool_raise_if import *
from base_aux.base_values.m3_exceptions import Exc__Expected


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        # SINGLE ---
        ((False,), [False, False, True, True]),
        ((None,), [False, False, True, True]),
        ((True,), [True, True, False, False]),

        ((0,), [False, False, True, True]),
        ((1,), [True, True, False, False]),

        ((Exception, ), [False, False, True, True]),

        ((LAMBDA_FALSE,), [False, False, True, True]),
        ((LAMBDA_NONE,), [False, False, True, True]),
        ((LAMBDA_TRUE,), [True, True, False, False]),
        ((LAMBDA_RAISE,), [False, False, True, True]),
        ((LAMBDA_EXC,), [False, False, True, True]),

        # SEVERAL ---
        ((0, 1,), [False, True, False, True]),
        ((LAMBDA_RAISE, 1,), [False, True, False, True]),
    ]
)
def test__RaiseIf(args, _EXPECTED):
    Lambda(ArgsBoolIf_AllTrue(*args).resolve).check_expected__assert(_EXPECTED[0])
    Lambda(ArgsBoolIf_AnyTrue(*args).resolve).check_expected__assert(_EXPECTED[1])
    Lambda(ArgsBoolIf_AllFalse(*args).resolve).check_expected__assert(_EXPECTED[2])
    Lambda(ArgsBoolIf_AnyFalse(*args).resolve).check_expected__assert(_EXPECTED[3])

    Lambda(ArgsRaiseIf_AllTrue(*args).resolve).check_expected__assert(Exc__Expected if _EXPECTED[0] else False)
    Lambda(ArgsRaiseIf_AnyTrue(*args).resolve).check_expected__assert(Exc__Expected if _EXPECTED[1] else False)
    Lambda(ArgsRaiseIf_AllFalse(*args).resolve).check_expected__assert(Exc__Expected if _EXPECTED[2] else False)
    Lambda(ArgsRaiseIf_AnyFalse(*args).resolve).check_expected__assert(Exc__Expected if _EXPECTED[3] else False)


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m1_init2_annots2_by_types import *

from base_aux.base_types.m2_info import *


# =====================================================================================================================
class Victim1(NestInit_AnnotsByTypes_All):
    NONE: None
    BOOL: bool
    INT: int
    FLOAT: float
    STR: str
    BYTES: bytes
    TUPLE: tuple
    LIST: list
    SET: set
    DICT: dict

    OPTIONAL: Optional
    OPTIONAL_BOOL: Optional[bool]

    # UNION: Union
    UNION_BOOL_INT: Union[bool, int]


victim1 = Victim1()


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="arg, _EXPECTED",
    argvalues=[
        ("NONE", None),
        ("BOOL", False),
        ("INT", 0),
        ("FLOAT", 0.0),
        ("STR", ""),
        ("BYTES", b""),
        ("TUPLE", ()),
        ("LIST", []),
        ("SET", set()),
        ("DICT", dict()),

        ("OPTIONAL", None),
        ("OPTIONAL_BOOL", None),
        ("UNION_BOOL_INT", False),

        ("NEVER", Exception),
    ]
)
def test__all(arg, _EXPECTED):
    Lambda(getattr, victim1, arg).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Victim2(NestInit_AnnotsByTypes_NotExisted):
    NOTEXIST: int
    EXIST: int = 100


victim2 = Victim2()

@pytest.mark.parametrize(
    argnames="arg, _EXPECTED",
    argvalues=[
        ("NOTEXIST", 0),
        ("EXIST", 100),

        ("NEVER", Exception),
    ]
)
def test__not_existed(arg, _EXPECTED):
    Lambda(getattr, victim2, arg).check_expected__assert(_EXPECTED)


# =====================================================================================================================

if __name__ == "__main__":
    # print(AttrAux_AnnotsAll(victim).dump__dict_types())
    # print(AttrAux_AnnotsAll(victim).dump__dict_values())

    ObjectInfo(victim1.__annotations__["UNION_BOOL_INT"]).print()


# =====================================================================================================================



from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_values.m3_exceptions import Exc__Expected
from base_aux.aux_eq.m4_eq_valid_chain import *


# =====================================================================================================================
class Base_EqRaiseIf(Base_EqValidChain):
    """
    GOAL
    ----
    if other value validated with any variant - raise! otherwise return None

    SPECIALLY CREATED FOR
    ---------------------
    replace ClsInstPrefix Raise_If__*
    """
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE
    MSG: str = None

    def __init__(self, *args, msg: str = None, **kwargs) -> None:
        if msg is not None:
            self.MSG = msg
        super().__init__(*args, **kwargs)

    def resolve(self, other_draft: Any, *other_args, **other_kwargs) -> None | NoReturn:
        validated = super().resolve(other_draft, *other_args, **other_kwargs)
        if validated:
            if self.MSG is not None:
                msg = str(self.MSG)
            else:
                msg = str(self)
            raise Exc__Expected(msg)


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqRaiseIf_Any(Base_EqRaiseIf):
    """
    NOTE
    ----
    for Raise it is MOST USEFUL
    """
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE


@final
class EqRaiseIf_All(Base_EqRaiseIf):
    """
    NOTE
    ----
    for Chain it is LESS USEFUL
    but created just to keep mirror
    """
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE


# =====================================================================================================================
if __name__ == "__main__":
    pass

    1 == EqRaiseIf_Any(0)

    try:
        1 == EqRaiseIf_Any(1)
    except:
        assert True
    else:
        assert False


# =====================================================================================================================



from base_aux.base_nest_dunders.m1_init0_annots_required import *
from base_aux.base_nest_dunders.m1_init0_reinit1_mutable_attrs import *
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *
from base_aux.base_nest_dunders.m2_str2_attrs import *
from base_aux.base_nest_dunders.m6_eq1_attrs import *
from base_aux.base_nest_dunders.m8_contains_attrs import *
from base_aux.base_nest_dunders.m8_len_attrs import *


# =====================================================================================================================
class Base_AttrKit(
    NestInit_AnnotsAttr_ByArgsKwargs,

    NestGAI_AnnotAttrIC,
    NestInit_MutableClsValues,
    NestInit_AnnotsRequired,

    NestEq_AttrsNotHidden,
    NestStR_AttrsNotHidden,
    NestLen_AttrNotHidden,
    NestContains_AttrIcNotHidden,
):     # TODO: decide to delete! use only dynamic?? - NO! keep it!!!
    """
    SAME AS - 1=ATTRS
    -----------------
    https://www.attrs.org/en/stable/examples.html
    yes! but more simple! and not so complicated! clear one way logic!

    SAME AS - 2=NAMEDTUPLE
    ----------------------

    GOAL
    ----
    just show that child is a kit
    1/ attrs need to init by args/kwargs
    2/ all annotated - must set!
    3/ IgnoreCase applied!

    NOTE
    ----
    !/ DONT USE DIRECTLY! use AttrKit_Blank instead! direct usage acceptable only for isinstance checking!
    1/ used in final CHILDs
    2/ basically used for static values like parsed from ini/json files

    SPECIALLY CREATED FOR
    ---------------------
    create special final kits like AttrKit_AuthNamePwd

    OLD docstr
    =======================
        GOAL
        ----
        1/ generate object with exact attrs values by Kwargs like template
        2/ for further comparing by Eq
        3/ all callables will resolve as Exc
    """
    def _redefine_nones(self, *args, **kwargs) -> None:
        """
        GOAL
        ----
        after created instance you can reapply defaults
        so if values keep None and you vant ro reinit it - just pass nes values!

        SPECIALLY CREATED FOR
        ---------------------
        Base_ReAttempts when you want to pass attempts by Rexp-patterns (with some nones) and define default values later in future methods
        """
        # TODO: finish!
        for index, value in enumerate(args):
            pass

        for name, value in kwargs.items():
            pass


# =====================================================================================================================
@final
class AttrKit_Blank(Base_AttrKit):
    """
    GOAL
    ----
    jast show that you can create any kwargs kit without raising (when check annots required)
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
@final
class AttrKit_AuthNamePwd(Base_AttrKit):
    NAME: str
    PWD: str


@final
class AttrKit_AuthTgBot(Base_AttrKit):
    LINK_ID: str = None     # @mybot20230913
    NAME: str = None        # MyBotPublicName
    TOKEN: str


@final
class AttrKit_AuthServer(Base_AttrKit):
    NAME: str
    PWD: str
    SERVER: str


@final
class AttrKit_AddrPort(Base_AttrKit):
    """class for keeping connection parameters/settings for exact smtp server

    :ivar ADDR: smtp server address like "smtp.mail.ru"
    :ivar PORT: smtp server port like 465
    """
    ADDR: str
    PORT: int


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================



import pytest
import pathlib

from base_aux.base_lambdas.m1_lambda import *
from base_aux.path1_dir.m1_dirpath import *


# =====================================================================================================================
CWD = pathlib.Path().cwd()


# =====================================================================================================================
class Test_Dirpath:
    def test__resolve(self):
        assert Resolve_DirPath().resolve() == CWD
        assert Resolve_DirPath("").resolve() != CWD

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            # dirs ---------
            (None, CWD),
            ('', pathlib.Path(".")),
            ('.', pathlib.Path(".")),
            ('hello', pathlib.Path("hello")),

            # files --------
            (CWD.joinpath('fileNotExisted.txt'), CWD),
            (CWD.joinpath('txt'), CWD.joinpath('txt')),
            (CWD.joinpath('.txt'), CWD),
            (CWD.joinpath('fileNotExisted.'), CWD.joinpath('fileNotExisted.')),
            (CWD.joinpath('fileNotExisted.txt1'), CWD.joinpath('fileNotExisted.txt1')),
        ]
    )
    def test__1(self, source, _EXPECTED):
        func_link = Resolve_DirPath(source).resolve
        Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================



from base_aux.base_nest_dunders.m1_init1_args_kwargs import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_values.m3_exceptions import Exc__Expected
from base_aux.base_nest_dunders.m3_calls import *
from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
class Base_ArgsBoolIf(NestInit_Args_Implicit, NestCall_Resolve):
    """
    GOAL
    ----
    check args (callables acceptable)
    and raise if True
    """
    ARGS: Any | Callable[..., Any] = ()
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE

    RAISE_INSTEAD_TRUE: bool = None  # RAISE_INSTEAD_TRUE
    RAISE_ON_INIT: bool = None

    def __init__(self, *args, _iresult_cumulate: EnumAdj_BoolCumulate = None, _raise_instead_true: bool = None, raise_on_init: bool = None, **kwargs) -> None | NoReturn:
        super().__init__(*args, **kwargs)

        if _iresult_cumulate is not None:
            self.IRESULT_CUMULATE = _iresult_cumulate

        if _raise_instead_true is not None:
            self.RAISE_INSTEAD_TRUE = bool(_raise_instead_true)

        if raise_on_init is not None:
            self.RAISE_ON_INIT = bool(raise_on_init)

        if self.RAISE_ON_INIT and self.RAISE_INSTEAD_TRUE:
            self.resolve()

    def resolve(self) -> bool | NoReturn:
        results_all = []
        for arg_source in self.ARGS:
            arg_result = Lambda(arg_source).resolve__bool()
            results_all.append(arg_result)

            if self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_TRUE:
                if arg_result:
                    continue
                else:
                    return False

            elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_FALSE:
                if not arg_result:
                    continue
                else:
                    return False

            elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_TRUE:
                if arg_result:
                    if self.RAISE_INSTEAD_TRUE:
                        msg = f"{arg_source=}/{arg_result=}//{results_all=}"
                        raise Exc__Expected(msg)
                    else:
                        return True

            elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_FALSE:
                if not arg_result:
                    if self.RAISE_INSTEAD_TRUE:
                        msg = f"{arg_source=}/{arg_result=}//{results_all=}"
                        raise Exc__Expected(msg)
                    else:
                        return True

        # FINAL ---------
        if self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_TRUE or self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_FALSE:
            return False
        else:   # self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_TRUE or self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_FALSE
            if self.RAISE_INSTEAD_TRUE:
                msg = f"{results_all=}"
                raise Exc__Expected(msg)
            else:
                return True


# =====================================================================================================================
class ArgsBoolIf_AllTrue(Base_ArgsBoolIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE


class ArgsBoolIf_AnyTrue(Base_ArgsBoolIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE


class ArgsBoolIf_AllFalse(Base_ArgsBoolIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_FALSE


class ArgsBoolIf_AnyFalse(Base_ArgsBoolIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_FALSE


# =====================================================================================================================
class Base_ArgsRaiseIf(Base_ArgsBoolIf):
    RAISE_INSTEAD_TRUE: bool = True


# ---------------------------------------------------------------------------------------------------------------------
class ArgsRaiseIf_AllTrue(Base_ArgsRaiseIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE


class ArgsRaiseIf_AnyTrue(Base_ArgsRaiseIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_TRUE


class ArgsRaiseIf_AllFalse(Base_ArgsRaiseIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_FALSE


class ArgsRaiseIf_AnyFalse(Base_ArgsRaiseIf):
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ANY_FALSE


# ---------------------------------------------------------------------------------------------------------------------
class ArgsRaiseIfOnInit_AllTrue(ArgsRaiseIf_AllTrue):
    RAISE_ON_INIT: bool = True


class ArgsRaiseIfOnInit_AnyTrue(ArgsRaiseIf_AnyTrue):
    RAISE_ON_INIT: bool = True


class ArgsRaiseIfOnInit_AllFalse(ArgsRaiseIf_AllFalse):
    RAISE_ON_INIT: bool = True


class ArgsRaiseIfOnInit_AnyFalse(ArgsRaiseIf_AnyFalse):
    RAISE_ON_INIT: bool = True


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class NestGa_Self:
    """
    GOAL
    ----
    in tests when we need some object like
        stand_schema.device_table.count_columns()
    but we dont want to create all levels!

    SPECIALLY CREATED FOR
    ---------------------
    BaseTgcPhases_Case tests:
    we need to pass stand_schema,
    but used only stand_schema.device_table.count_columns()

    NOTE
    ----
    it is only a part of realisation!
    1. create expected methods in child!
    2. create expected attrs in init!

    EXAMPLE
    -------
    1. use any chain if not exists ga
        class Victim(NestGa_Self): ...
        victim = Victim()
        assert victim == victim.hello.world

    2. use as nested
        class Victim(NestGa_Self):
            attr=1
            def meth(self, param=None):
                return param
        victim = Victim()
        assert victim.attr == victim.hello.world.attr
        assert victim.meth() == victim.hello.world.meth()
    """
    def __getattr__(self, item: str) -> Self:
        return self


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.aux_eq.m4_eq_valid_chain import *

from base_aux.aux_eq.m5_eq_raise_if import *
from base_aux.base_values.m3_exceptions import Exc__Expected


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, other, _EXPECTED",
    argvalues=[
        ((1,), LAMBDA_RAISE, [False, False]),
        ((1,), 1, [True, True]),
        ((1, 2), 1, [False, True]),
        ((1, 2), 0, [False, False]),

        ((EqValid_Raise(), EqValid_NotRaise(), ), LAMBDA_RAISE, [False, True]),
        ((EqValid_NotRaise(), EqValid_Raise(), ), LAMBDA_RAISE, [False, True]),

        ((EqValid_NotRaise(), ), LAMBDA_RAISE, [False, False]),
        ((EqValid_NotRaise(), ), 1, [True, True]),
        ((EqValid_NotRaise(), EqValid_Raise()), 1, [False, True]),

        ((EqValid_NotRaise(), EqValid_GE(1)), 1, [True, True]),
        ((EqValid_NotRaise(), EqValid_GE(100)), 1, [False, True]),
        ((EqValid_NotRaise(), EqValid_GE(100, _iresult_reverse=True)), 1, [True, True]),
    ]
)
def test___EqValidator(args, other, _EXPECTED):
    Lambda(EqValidChain_All(*args) == other).check_expected__assert(_EXPECTED[0])
    Lambda(EqValidChain_Any(*args) == other).check_expected__assert(_EXPECTED[1])

    Lambda(lambda: EqRaiseIf_All(*args) == other).check_expected__assert(Exc__Expected if _EXPECTED[0] else None)
    Lambda(lambda: EqRaiseIf_Any(*args) == other).check_expected__assert(Exc__Expected if _EXPECTED[1] else None)


# =====================================================================================================================



from typing import *
import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_attr.m3_ga1_prefix_1_inst import NestGa_Prefix


# =====================================================================================================================
class Victim(NestGa_Prefix):
    GETATTR_PREFIXES = ["bool__", ]
    TRUE = True
    NONE = None

    def bool__(self, value: Any = None) -> bool | NoReturn:
        return bool(value)

    def meth_true(self):
        return True

    def meth_echo(self, value):
        return value


victim = Victim()


# =====================================================================================================================
def test__anycase_attr():
    assert victim.TRUE == True
    assert victim.true == True
    assert victim.NONE == None
    assert victim.none == None


def test__1():
    assert victim.bool__() == False
    assert victim.bool__(True) == True

    assert victim.bool__true() == True
    assert victim.bool__meth_true() == True

    try:
        victim.bool__meth_echo()
    except:
        assert True
    else:
        assert False

    assert victim.bool__meth_echo(1) == True
    assert victim.bool__meth_echo(0) == False
    assert victim.bool__meth_echo(True) == True
    assert victim.bool__meth_echo(False) == False


def test__anycase_meth():
    assert victim.BOOL__() == False
    assert victim.BOOL__(True) == True

    assert victim.BOOL__TRUE() == True
    assert victim.BOOL__METH_TRUE() == True

# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="meth, args, _EXPECTED",
    argvalues=[
        ("bool__", (),  False),
        ("BOOL__", (), False),

        ("bool11__true111", (), Exception),
        ("bool__true111", (), Exception),
        ("bool__true", (), True),
        ("BOOL__TRUE", (), True),
        ("BOOL__NONE", (), False),

        ("BOOL__meth_true", (), True),
        ("BOOL__meth_true", (123, ), Exception),

        ("BOOL__meth_echo", (), Exception),
        ("BOOL__meth_echo", (1, ), True),
        ("BOOL__meth_echo", ("123", ), True),

    ]
)
def test__batch(meth, args, _EXPECTED):
    func_link = lambda *_args: getattr(victim, meth)(*_args)
    Lambda(func_link, *args).check_expected__assert(_EXPECTED)


# =====================================================================================================================



import pytest
from base_aux.base_lambdas.m1_lambda import *

from base_aux.base_values.m4_primitives import *
from base_aux.aux_argskwargs.m1_argskwargs import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_nest_dunders.m7_cmp import *


# =====================================================================================================================
class Cls(NestCmp_GLET_Any):
    def __init__(self, value):
        self.VALUE = value

    def __cmp__(self, other):
        other = Cls(other)
        if self.VALUE == other.VALUE:
            return 0
        if self.VALUE > other.VALUE:
            return 1
        if self.VALUE < other.VALUE:
            return -1


def test____LE__():
    func_link = lambda result: result == 1
    Lambda(func_link, Cls(1)).check_expected__assert(True)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="func_link, args, _EXPECTED_RAISED",
    argvalues=[
        # Special Values -------
        # (NoValue, (), False),   # CANT CHECK NoValue))))

        # not callable ------------
        (True, (), False),
        (True, (111, ), False),
        (False, (), False),

        # callable ------------
        (LAMBDA_ECHO, (), True),
        (LAMBDA_ECHO, (None, ), False),
        (LAMBDA_ECHO, (True, ), False),
        (lambda value: value, (), True),

        # TYPES -------
        (int, (), False),
        (1, (), False),
        (Exception, (), False),
        (Exception(), (), False),
        (LAMBDA_RAISE, (), True),
    ]
)
def test__check_all_meth__simple(
        func_link: Any | Callable,
        args: Any,

        _EXPECTED_RAISED: bool,
):
    assert Lambda(func_link, *args).check_raised__bool(_EXPECTED_RAISED) is True
    assert Lambda(func_link, *args).check_no_raised__bool(not _EXPECTED_RAISED) is True

    if _EXPECTED_RAISED:
        assert Lambda(func_link, *args).check_raised__bool() is True
        assert Lambda(func_link, *args).check_raised__bool(True) is True
        assert Lambda(func_link, *args).check_raised__bool(False) is False

        assert Lambda(func_link, *args).check_no_raised__bool() is False
        assert Lambda(func_link, *args).check_no_raised__bool(True) is False
        assert Lambda(func_link, *args).check_no_raised__bool(False) is True

        # assert RAISED-----------------
        assert Lambda(func_link, *args).check_raised__assert() is None
        assert Lambda(func_link, *args).check_raised__assert(True) is None
        try:
            Lambda(func_link, *args).check_raised__assert(False)
        except:
            assert True
        else:
            assert False

        # assert NoRAISED-----------------
        try:
            Lambda(func_link, *args).check_no_raised__assert()
        except:
            assert True
        else:
            assert False
        try:
            Lambda(func_link, *args).check_no_raised__assert(True)
        except:
            assert True
        else:
            assert False
        assert Lambda(func_link, *args).check_no_raised__assert(False) is None

    else:
        assert Lambda(func_link, *args).check_raised__bool() is False
        assert Lambda(func_link, *args).check_raised__bool(True) is False
        assert Lambda(func_link, *args).check_raised__bool(False) is True

        assert Lambda(func_link, *args).check_no_raised__bool() is True
        assert Lambda(func_link, *args).check_no_raised__bool(True) is True
        assert Lambda(func_link, *args).check_no_raised__bool(False) is False

        # assert RAISED-----------------
        try:
            Lambda(func_link, *args).check_raised__assert()
        except:
            assert True
        else:
            assert False
        try:
            Lambda(func_link, *args).check_raised__assert(True)
        except:
            assert True
        else:
            assert False
        assert Lambda(func_link, *args).check_raised__assert(False) is None

        # assert NoRAISED-----------------
        assert Lambda(func_link, *args).check_no_raised__assert() is None
        assert Lambda(func_link, *args).check_no_raised__assert(True) is None
        try:
            Lambda(func_link, *args).check_no_raised__assert(False)
        except:
            assert True
        else:
            assert False


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="func_link, args, kwargs, _EXPECTED, _Expected_pytestResult, _EXP_raised",
    argvalues=[
        # Special Values -------
        # (NoValue, (), {}, NoValue, True, False),   # CANT CHECK NoValue))))

        # not callable ------------
        (True, (), {}, True, True, False),

        (True, (111, ), {"111": 222}, True, True, False),
        (True, (111, ), {"111": 222}, False, False, False),

        (False, (), {}, True, False, False),

        # callable ------------
        (LAMBDA_ECHO, (), {}, True, False, True),

        (LAMBDA_ECHO, (None, ), {}, True, False, False),
        (LAMBDA_ECHO, (None, ), {}, None, True, False),
        (LAMBDA_ECHO, (True, ), {}, True, True, False),
        (LAMBDA_ECHO, (True, ), {}, True, True, False),
        (lambda value: value, (), {"value": True}, True, True, False),
        (lambda value: value, (), {"value": None}, True, False, False),

        # TYPES -------
        (int, (), {}, int, True, False),
        (1, (), {}, int, True, False),
        (1, (), {}, float, False, False),

        (1, (), {}, Exception, False, False),
        (Exception, (), {}, Exception, True, False),
        (Exception(), (), {}, Exception, True, False),

        (LAMBDA_RAISE, (), {}, Exception, True, True),
    ]
)
def test__check_all_meth__complex(
        func_link: Any | Callable,
        args: Any,
        kwargs: Any,
        _EXPECTED: Any | type[Any],
        _Expected_pytestResult: bool,
        _EXP_raised: bool
):
    # check_RAISED ----------------------------
    assert Lambda(func_link, *args, **kwargs).check_raised__bool() is _EXP_raised
    assert Lambda(func_link, *args, **kwargs).check_no_raised__bool() is not _EXP_raised

    if _EXP_raised:
        # ---------------
        assert Lambda(func_link, *args, **kwargs).check_raised__assert() is None

        # ---------------
        try:
            Lambda(func_link, *args, **kwargs).check_no_raised__assert()
        except:
            assert True
        else:
            assert False

    else:
        # ---------------
        assert Lambda(func_link, *args, **kwargs).check_no_raised__assert() is None

        # ---------------
        try:
            Lambda(func_link, *args, **kwargs).check_raised__assert()
        except:
            assert True
        else:
            assert False

    # check_EXPECTED --------------------------
    assert Lambda(func_link, *args, **kwargs).check_expected__bool(_EXPECTED) == _Expected_pytestResult
    try:
        Lambda(func_link, *args, **kwargs).check_expected__assert(_EXPECTED)
    except:
        assert not _Expected_pytestResult
    else:
        assert _Expected_pytestResult
        # assert result == _Expected_pytestResult


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        ((), {}, []),
        ((None, ), {}, [None, ]),
        ((1, ), {}, [1, ]),
        ((1, 1), {}, [1, 1]),

        ((1, 1), {}, [1, 1]),
        ((1, 1), {"2": 22}, [1, 1, "2"]),
        ((1, 1), {"2": 22, "3": 33}, [1, 1, "2", "3"]),
    ]
)
def test__func_list_direct(args, kwargs, _EXPECTED):
    Lambda(LAMBDA_LIST_KEYS, *args, **kwargs).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        ((), {}, []),
        ((None, ), {}, [None, ]),
        ((1, ), {}, [1, ]),
        ((1, 1), {}, [1, 1]),

        ((1, 1), {}, [1, 1]),
        ((1, 1), {"2": 22}, [1, 1, 22]),
        ((1, 1), {"2": 22, "3": 33}, [1, 1, 22, 33]),
    ]
)
def test__func_list_values(args, kwargs, _EXPECTED):
    Lambda(LAMBDA_LIST_VALUES, *args, **kwargs).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        ((), {}, {}),
        ((None, ), {}, {None: None}),
        ((1, ), {}, {1: None}),
        ((1, 1), {}, {1: None}),

        ((1, 1), {}, {1: None}),
        ((1, 1), {"2": 22}, {1: None, "2": 22}),
        ((1, 1), {"2": 22, "3": 33}, {1: None, "2": 22, "3": 33}),
    ]
)
def test__func_dict(args, kwargs, _EXPECTED):
    Lambda(LAMBDA_DICT_KEYS, *args, **kwargs).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        ((), {}, True),
        ((None, ), {}, False),
        ((1, ), {}, True),
        ((1, 1), {}, True),

        ((1, 1), {}, True),
        ((1, 1), {"2": 22}, True),
        ((1, 1), {"2": 22, "3": 33}, True),

        ((1, 1), {"2": 22, "3": None}, False),
    ]
)
def test__func_all(args, kwargs, _EXPECTED):
    Lambda(LAMBDA_ALL_VALUES, *args, **kwargs).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED",
    argvalues=[
        ((), {}, False),
        ((None, ), {}, False),
        ((1, ), {}, True),
        ((1, 1), {}, True),

        ((1, 1), {}, True),
        ((1, 1), {"2": 22}, True),
        ((1, 1), {"2": 22, "3": 33}, True),

        ((1, 1), {"2": 22, "3": None}, True),
        ((1, None), {"2": 22, "3": None}, True),
        ((None, None), {"2": True, "3": None}, True),
        ((None, None), {"2": False, "3": None}, False),

        (Args(None, None), {"2": False, "3": None}, False),
    ]
)
def test__func_any(args, kwargs, _EXPECTED):
    Lambda(LAMBDA_ANY_VALUES, *args, **kwargs).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, other, _EXPECTED",
    argvalues=[
        ("11.688889V", EqValid_Regexp(r"\d+[.,]?\d*V"), True),
        (INST_EQ_TRUE, INST_EQ_TRUE, True),
        (INST_EQ_TRUE, INST_EQ_FALSE, True),
        (INST_EQ_FALSE, INST_EQ_TRUE, True),
        (INST_EQ_FALSE, INST_EQ_FALSE, False),
    ]
)
def test__EQ(source, other, _EXPECTED):
    assert Lambda(source).check_expected__bool(other) == _EXPECTED


# =====================================================================================================================



from typing import *


# TEST STUFF ==========================================================================================================
class Row_:
    NAME: str = "row"
    SKIP: Optional[bool] = None
    result: Optional[bool] = None

    def __init__(self, name: Optional[Any] = None):
        if name:
            self.NAME = str(name)


class Dev_:
    NAME: str = "dev"
    SKIP: Optional[bool] = None
    result: Optional[bool] = None

    def __init__(self, name: Optional[Any] = None):
        if name:
            self.NAME = str(name)

    def SKIP_reverse(self) -> None:
        self.SKIP = not bool(self.SKIP)


class Data_:
    ROWS: list[Row_] = None
    DEVS: list[Dev_] = None

    def __init__(self, rows: list[Row_], devs: list[Dev_]):
        self.ROWS = list(rows)
        self.DEVS = list(devs)


# =====================================================================================================================



import pytest
from base_aux.buses.m0_history import *


# =====================================================================================================================
class Test__HistoryIO:
    VICTIM: type[HistoryIO] = type("Victim", (HistoryIO,), {})

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        self.VICTIM = type("Victim", (HistoryIO,), {})

    def teardown_method(self, method):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__add_list_asdict(self):
        victim: HistoryIO = self.VICTIM()
        assert victim.history == []
        assert victim.as_dict() == {}

        victim.add_input("in1")
        assert victim.history == [("in1", [])]
        assert victim.list_input() == ["in1", ]
        assert victim.list_output() == []
        assert victim.as_dict() == {"in1": []}

        victim.add_input("in2")
        assert victim.history == [("in1", []), ("in2", [])]
        assert victim.list_input() == ["in1", "in2"]
        assert victim.list_output() == []
        assert victim.as_dict() == {"in1": [], "in2": []}

        victim.add_output("out2")
        assert victim.history == [("in1", []), ("in2", ["out2", ])]
        assert victim.list_input() == ["in1", "in2"]
        assert victim.list_output() == ["out2", ]

        victim.add_output("out22")
        assert victim.history == [("in1", []), ("in2", ["out2", "out22", ])]
        assert victim.list_input() == ["in1", "in2"]
        assert victim.list_output() == ["out2", "out22", ]

        victim.add_output(["out222", "out2222"])
        assert victim.history == [("in1", []), ("in2", ["out2", "out22", "out222", "out2222", ])]
        assert victim.list_input() == ["in1", "in2"]
        assert victim.list_output() == ["out2", "out22", "out222", "out2222", ]

        victim.print_io()

    def test__add_io__list_last(self):
        victim: HistoryIO = self.VICTIM()
        assert victim.history == []
        assert victim.last_input == ""
        assert victim.last_output == ""

        victim.add_io("in1", "out1")
        assert victim.history == [("in1", ["out1", ]), ]
        assert victim.list_input() == ["in1", ]
        assert victim.list_output() == ["out1", ]
        assert victim.last_input == "in1"
        assert victim.last_output == "out1"

        victim.add_io("in2", ["out2", "out22", ])
        assert victim.history == [("in1", ["out1", ]), ("in2", ["out2", "out22", ]), ]
        assert victim.list_input() == ["in1", "in2"]
        assert victim.list_output() == ["out1", "out2", "out22", ]
        assert victim.last_input == "in2"
        assert victim.last_output == "out22"

    def test__add_history(self):
        victim: HistoryIO = self.VICTIM()
        assert victim.history == []

        history = HistoryIO()
        history.add_io("in1", "out1")

        victim.add_history(history)
        assert victim.history == [("in1", ["out1", ]), ]

    def test__first_output(self):
        victim: HistoryIO = self.VICTIM()
        assert victim.history == []

        victim.add_output("out0")
        assert victim.history == [("", ["out0", ])]
        assert victim.list_input() == ["", ]
        assert victim.list_output() == ["out0"]

        victim.clear()
        assert victim.history == []


# =====================================================================================================================



from typing import *
import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_dict.m2_dict_ic import *
from base_aux.aux_dict.m3_dict_ga import *


# # =====================================================================================================================
# @pytest.mark.parametrize(
#     argnames="source, key, keys_all_str, _EXPECTED",
#     argvalues=[
#         (dict(attr1=1), "attr1", True, [None, None]),
#         (dict(attr1=1), "ATTR1", True, [None, None]),
#         (dict(ATTR1=1), "ATTR1", True, [None, None]),
#         (dict(attr1=1), "hello", True, [None, Exception]),
#
#         (dict(attr1=1), 0, False, [None, Exception]),
#         ({1:1}, 0, False, [None, Exception]),
#         ({1:1}, 1, False, [None, None]),
#     ]
# )
# @pytest.mark.parametrize(argnames="VictimClsPair", argvalues=[(DictIcKeys, DictIc_LockedKeys), (DictIcKeys_Ga, DictIc_LockedKeys_Ga)])
# def test__si_update(VictimClsPair, source, key, keys_all_str, _EXPECTED):
#     # -------------------------------------------------
#     victim = VictimClsPair[0](source)
#     Lambda(victim.update({key: 11})).expect__check_assert(_EXPECTED[0])
#
#
# # =====================================================================================================================



from typing import *
import pathlib
import csv
import threading

import requests
from bs4 import BeautifulSoup

from base_aux.alerts.m2_select import *
from base_aux.alerts.m1_alert0_base import *


# =====================================================================================================================
# TODO: separate ClsBASE! see Imap!


# =====================================================================================================================
class TagAddressChain(NamedTuple):
    """
    GOAL
    ----
    create several chains to point to one tag.

    structure which used as one step of full chain for finding Tag.
    all this params directly will be passed into function Tag.find_all!
    There is no variants - only full values!

    :ivar NAME: tag name for finding
    :ivar ATTRS: dict with expected attr names and exact values
        {"class": "donor-svetofor-restyle"}
    :ivar TEXT: text in expected tag
        None - if not specified
    :ivar INDEX: in case of found several tags - get the exact index of them

    EXAMPLES
    --------
    see in monitor implementations
        TagAddressChain("table", {"class": "donor-svetofor-restyle"}, None, 0),
    """
    NAME: str
    ATTRS: dict[str, str]
    TEXT: Optional[str]
    INDEX: int


class Alert_MonitorUrlTag(AlertSelect.TELEGRAM_DEF):
    """
    GOAL
    ----
    Just created exact class for Alerts!
    """
    pass


# =====================================================================================================================
class MonitorUrlTag(threading.Thread):
    """base class for final monitors!
    monitoring on URL some value.
    if found new value - remember it and send mail alert!

    :ivar URL: monitoring url
    :ivar TIMEOUT_REQUEST: timeout in request for html data
    :ivar INTERVAL: monitoring interval
    :ivar TAG_CHAINS: full chain to find exact tag
    :ivar TAG_GET_ATTR: which text get from found tag,
        None - from tag context,
        stringValue - from exact attribute name
    :ivar ALERT: object which will send the alerts
    :ivar DIRPATH: path to directory where will be saved history values
    :ivar CSV_DELIMITER: delimeter for history csvFile

    :ivar _source_data: full html for url
    :ivar _tag_found_last_chain: temporary kept last found tag object by one Chain
    :ivar value_last: last extracted value, jast for ability to compare alert state
    :ivar value_prelast: prelast extracted value

    :ivar msg: message (may be accumulated through process)
    """
    # SETTINGS -------------------------------
    URL: str = None
    TIMEOUT_REQUEST: int = 10
    INTERVAL: int = 1 * 60 * 60
    TAG_CHAINS: list[TagAddressChain] = []
    TAG_GET_ATTR: Optional[str] = None
    ALERT: type[Base_Alert] = Alert_MonitorUrlTag

    DIRPATH: pathlib.Path = pathlib.Path("USERDATA")
    CSV_DELIMITER: str = ";"

    # internal ----------------------------------
    _source_data: str = ""
    _tag_found_last_chain: Optional[BeautifulSoup] = None
    value_last: str = None
    value_prelast: str = None
    msg: str = ""

    def __init__(self):
        super().__init__(daemon=False)

        self.DIRPATH = pathlib.Path(self.DIRPATH)
        self.FILEPATH = self.DIRPATH.joinpath(f"{self.NAME}.csv")

        self.DIRPATH.mkdir(exist_ok=True)
        if self.FILEPATH.exists():
            self.value_last__load()
        else:
            self.FILEPATH.touch(exist_ok=True)

        self.start()

    @property
    def NAME(self):
        """class name
        """
        return self.__class__.__name__

    def value_last__load(self) -> None:
        """load last value from history
        """
        if not self.FILEPATH.exists():
            return

        with open(self.FILEPATH, "rt", newline='') as ofilepath:
            reader = csv.reader(ofilepath, delimiter=self.CSV_DELIMITER)
            result = None
            for line_parsed in reader:
                result = line_parsed[1]
            self.value_last = self.value_prelast = result

    def value_last__save(self) -> None:
        """save last value from history
        """
        with open(self.FILEPATH, "a", newline='') as ofilepath:
            writer = csv.writer(ofilepath, delimiter=self.CSV_DELIMITER)
            writer.writerow([time.strftime("%Y.%m.%d %H:%M:%S"), self.value_last])

    def run(self):
        """MAIN function working in thread
        """
        while True:
            if self.alert_state__check():
                self.ALERT(self.msg)

            print(self.msg)
            time.sleep(self.INTERVAL)

    def alert_state__check(self) -> bool:
        """check alert state

        :returns:
            True - if need ALERT!
            there is only one way to return False - all funcs get true(correctly finished) and oldValue == newValue.
            Otherwise, need send email!!!
            if some func get false or finished incorrect - we will have smth in self.msg!
        """
        result = True
        if all([
            self.reinit_values(),
            self.source__load(),
            self.source__get_tag(),
            self.tag__get_value(),
            ]):

            if self.value_last != self.value_prelast:
                self.msg += f"DETECTED CHANGE[{self.value_prelast}->{self.value_last}]"
                self.value_last__save()
            else:
                self.msg += f"SameState[{self.value_prelast}->{self.value_last}]"
                result = False

        return result

    def reinit_values(self) -> True:
        """clear main values on new monitor cycle
        """
        self.msg = ""
        self._source_data = ""
        self._tag_found_last_chain = None
        return True

    def source__load(self) -> bool:
        """load url source data
        """
        self._source_data = ""
        try:
            response = requests.get(self.URL, timeout=self.TIMEOUT_REQUEST)
            self._source_data = response.text
            return True
        except Exception as exc:
            self.msg += f"LOST URL {exc!r}"

    def source__get_tag(self) -> Optional[bool]:
        """find tag from source and save tag object in attribute
        """
        if self._source_data:
            try:
                self._tag_found_last_chain = BeautifulSoup(markup=self._source_data, features='html.parser')
            except Exception as exc:
                self.msg += f"[CRITICAL] can't parse {self._source_data=}\n{exc!r}"
                return
        else:
            self.msg += f"[CRITICAL] empty {self._source_data=}"
            return

        try:
            for chain in self.TAG_CHAINS:
                tags = self._tag_found_last_chain.find_all(name=chain.NAME, attrs=chain.ATTRS, string=chain.TEXT, limit=chain.INDEX + 1)
                self._tag_found_last_chain = tags[chain.INDEX]
        except Exception as exc:
            self.msg += f"URL WAS CHANGED! can't find {chain=}\n{exc!r}"
            return

        return True

    def tag__get_value(self) -> Optional[bool]:
        """get resulting value from found tag
        """
        if not self._tag_found_last_chain:
            return

        self.value_prelast = self.value_last

        if self.TAG_GET_ATTR is None:
            self.value_last = self._tag_found_last_chain.string
        else:
            self.value_last = self._tag_found_last_chain[self.TAG_GET_ATTR][0]

        if isinstance(self.value_last, str):
            self.value_last = self.value_last.strip()
        return True


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *

from base_aux.valid.m3_valid_chains import *
from base_aux.valid.m2_valid_derivatives import *
from base_aux.base_values.m5_value_valid2_variants import *
from base_aux.base_values.m5_value_valid3_unit import *


# =====================================================================================================================
def test__1():
    assert ClsEq(1) == 1
    assert ClsEq(1) != 2

    assert 1 == ClsEq(1)
    assert 2 != ClsEq(1)


# ---------------------------------------------------------------------------------------------------------------------
def test__str():
    victim = Valid(True)
    victim.run()
    print(victim)

    victim = ValidChains([True, ])
    victim.run()
    print(victim)


# =====================================================================================================================
class Test__ValidTypes:
    # @classmethod
    # def setup_class(cls):
    #     pass
    #     cls.Victim = type("Victim", (ValueUnit,), {})
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="args, _EXPECTED",
        argvalues=[
            # BOOLS ---------------
            # direct TRUE
            ((0,), False),
            ((2,), False),  # careful about 1 comparing (assert 0 == False, assert 1 == True, assert 2 != True)
            (([],), False),
            (([None,],), False),
            (([1,],), False),

            ((0, True), False),
            ((2, True), False),
            (([], True), False),
            (([None, True],), False),
            (([1, ], True), False),

            # active BOOL
            ((0, bool), False),
            ((2, bool), True),
            (([], bool), False),
            (([None, ], bool), True),
            (([1, ], bool), True),

            # -----------------------
            ((LAMBDA_TRUE,), True),
            ((LAMBDA_TRUE, True), True),
            ((LAMBDA_TRUE, False), False),
            ((LAMBDA_TRUE, LAMBDA_TRUE), True),
            ((LAMBDA_TRUE, LAMBDA_FALSE), False),

            ((LAMBDA_NONE,), False),

            ((LAMBDA_FALSE,), False),
            ((LAMBDA_FALSE, False), True),
            ((LAMBDA_FALSE, LAMBDA_TRUE), True),
            ((LAMBDA_FALSE, LAMBDA_EXC), False),

            ((LAMBDA_EXC, True), False),
            ((LAMBDA_EXC, LAMBDA_TRUE), False),
            ((LAMBDA_EXC,), False),
            ((LAMBDA_EXC, LAMBDA_EXC), False),
            ((LAMBDA_EXC, Exception), True),

            ((True, None), True),
            ((lambda: True, None), True),

            ((True, lambda val: val is True), True),
            ((LAMBDA_TRUE, lambda val: val is True), True),

            ((lambda: 1, lambda val: 0 < val < 2), True),
            ((lambda: 1, lambda val: 0 < val < 1), False),

            ((lambda: "1", lambda val: 0 < val < 2), False),
            ((lambda: "1", lambda val: 0 < int(val) < 2), True),
            ((lambda: "1.0", lambda val: 0 < int(val) < 2), False),
            ((lambda: "1.0", lambda val: 0 < float(val) < 2), True),

            # ValueVariants --------------------------
            (("hello", ValueVariants(variants=["hello", 1])), True),
            (("1", ValueVariants(variants=["hello", 1])), True),
            ((1, ValueVariants(variants=["hello", 1])), True),
            (("0", ValueVariants(variants=["hello", 1])), False),
        ]
    )
    def test__validate__types(self, args, _EXPECTED):
        # DIRECT -------
        func_link = Valid(*args)
        Lambda(func_link).check_expected__assert(_EXPECTED)

        # REVERSE ------
        func_link = ValidReverse(*args)
        Lambda(func_link).check_expected__assert(not _EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, args, _EXPECTED",
        argvalues=[
            (1, (NoValue, lambda x: float(x) >= 1), True),
            ("1", (NoValue, lambda x: float(x) >= 1), True),
            ("0", (NoValue, lambda x: float(x) >= 1), False),
            ("hello", (NoValue, lambda x: float(x) >= 1), False),
        ]
    )
    def test__not_passed(self, source, args, _EXPECTED):
        # assert source == Valid(*valid_args)

        func_link = lambda *_args: source == Valid(*_args)
        Lambda(func_link, *args).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, args__value, kwargs__value, validate, _EXPECTED",
        argvalues=[
            # bool --------------------
            (0, (), {}, True, False),
            (0, (1,2,), {1:1}, True, Exception),

            (2, (), {}, bool, True),
            (2, (1,2,), {1:1}, bool, Exception),

            # VALUE --------------------
            (LAMBDA_LIST_VALUES, (1,2,), {}, [1,2], True),
            (LAMBDA_LIST_VALUES, (1,2,), {"1":11, }, [1,2,11], True),

            # ARG COLLECTION TYPES --------------------
            (LAMBDA_LIST_VALUES, 1, {}, [1, ], True),
            (LAMBDA_LIST_VALUES, ClsIterYield, {}, [ClsIterYield, ], True),
            (LAMBDA_LIST_VALUES, INST_ITER_YIELD, {}, [INST_ITER_YIELD, ], True),
            (LAMBDA_LIST_VALUES, INST_GEN, {}, [INST_GEN, ], True),
        ]
    )
    def test__value__args_kwargs(self, source, args__value, kwargs__value, validate, _EXPECTED):
        func_link = Valid(value_link=source, validate_link=validate, args__value=args__value, kwargs__value=kwargs__value)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, args__validate, kwargs__validate, validate, _EXPECTED",
        argvalues=[
            # bool --------------------
            (0, (), {}, True, False),
            (0, (1, 2,), {1: 1}, True, False),

            (2, (), {}, bool, True),
            (2, (1, 2,), {1: 1}, bool, Exception),

            # VALUE --------------------
            (LAMBDA_LIST_VALUES, (1, 2,), {}, [1, 2], False),
            (LAMBDA_LIST_VALUES, (1, 2,), {}, [], True),

            (LAMBDA_LIST_VALUES, (1, 2,), {"1": 11, }, [1, 2, 11], False),
            (LAMBDA_LIST_VALUES, (1, 2,), {"1": 11, }, [], True),

            # VALUE --------------------
            # (0, (1, 3,), {}, ValidAux_Obj.check_legt, False),    # FIXME: do smth!
            # (1, (1, 3,), {}, ValidAux_Obj.check_legt, True),
            # (2, (1, 3,), {}, ValidAux_Obj.check_legt, True),
            # (3, (1, 3,), {}, ValidAux_Obj.check_legt, False),
            # (4, (1, 3,), {}, ValidAux_Obj.check_legt, False),
            #
            # (0, (1, None,), {}, ValidAux_Obj.check_legt, False),
            # (1, (1, None,), {}, ValidAux_Obj.check_legt, True),
            # (2, (1, None,), {}, ValidAux_Obj.check_legt, True),
            #
            # (0, (None, 3,), {}, ValidAux_Obj.check_legt, True),
            # (1, (None, 3,), {}, ValidAux_Obj.check_legt, True),
            # (2, (None, 3,), {}, ValidAux_Obj.check_legt, True),
            # (3, (None, 3,), {}, ValidAux_Obj.check_legt, False),
        ]
    )
    def test__validate__args_kwargs(self, source, args__validate, kwargs__validate, validate, _EXPECTED):
        func_link = Valid(value_link=source, validate_link=validate, args__validate=args__validate, kwargs__validate=kwargs__validate)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="args",
        argvalues=[
            # BOOLING ---------------
            # direct TRUE
            (0,),
            (2,),  # careful about 1 comparing (assert 0 == False, assert 1 == True, assert 2 != True)
            ([],),
            ([None,],),
            ([1,],),

            (0, True),
            (2, True),
            (([], True)),
            ([None, True],),
            ([1, ], True),

            # active BOOL
            (0, bool),
            (2, bool),
            ([], bool),
            ([None, ], bool),
            ([1, ], bool),

            # -----------------------
            (LAMBDA_TRUE,),
            (LAMBDA_TRUE, True),
            (LAMBDA_TRUE, False),
            (LAMBDA_TRUE, LAMBDA_TRUE),
            (LAMBDA_TRUE, LAMBDA_FALSE),

            (LAMBDA_NONE,),

            (LAMBDA_FALSE,),
            (LAMBDA_FALSE, False),
            (LAMBDA_FALSE, LAMBDA_TRUE),
            (LAMBDA_FALSE, LAMBDA_EXC),

            (LAMBDA_EXC, True),
            (LAMBDA_EXC, LAMBDA_TRUE),
            (LAMBDA_EXC,),
            (LAMBDA_EXC, LAMBDA_EXC),
            (LAMBDA_EXC, Exception),

            (True, None),
            (lambda: True, None),

            (True, lambda val: val is True),
            (LAMBDA_TRUE, lambda val: val is True),

            (lambda: 1, lambda val: 0 < val < 2),
            (lambda: 1, lambda val: 0 < val < 1),

            (lambda: "1", lambda val: 0 < val < 2),
            (lambda: "1", lambda val: 0 < int(val) < 2),
            (lambda: "1.0", lambda val: 0 < int(val) < 2),
            (lambda: "1.0", lambda val: 0 < float(val) < 2),
        ]
    )
    def test__str(self, args):
        assert str(Valid(*args)) is not None

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, validate_link, retry, _EXPECTED",
        argvalues=[
            # bool --------------------
            ([True, False, False], list.pop, 0, False),
            ([True, False, False], list.pop, 1, False),
            ([True, False, False], list.pop, 2, True),
        ]
    )
    def test__retry(self, source, validate_link, retry, _EXPECTED):
        func_link = Valid(value_link=source, validate_link=validate_link, validate_retry=retry).run
        Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================



from typing import *

from base_aux.base_nest_dunders.m3_calls import *
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *


# =====================================================================================================================
class Url(NestInit_AnnotsAttr_ByArgsKwargs, NestCall_Resolve):
    PROTOCOL: str = "http"
    HOST: str = "localhost"
    PORT: int = 80
    ROUTE: str = ""

    _HIDE_PORT80: bool = None

    def resolve(
            self,
            protocol: Optional[str] = None,
            host: Optional[str] = None,
            port: Optional[int] = None,
            route: Optional[str | tuple[str|Any, ...]] = None,
    ) -> str:
        # ---------------------
        if protocol is None:
            protocol = self.PROTOCOL

        # ---------------------
        if host is None:
            host = self.HOST
        else:
            host = str(host)
        if host == "0.0.0.0":
            host = "localhost"

        # ---------------------
        if port is None:
            port = self.PORT

        # ---------------------
        if route is None:
            route = self.ROUTE

        if isinstance(route, (tuple, list)):
            route_new = ""
            for item in list(route):
                route_new += f"/{item}"
            route = route_new
        else:
            route = str(route)

        while route.startswith("/"):
            route = route[1:]

        # FINAL -------------------
        if self._HIDE_PORT80 and port == 80:
            result = f"{protocol}://{host}/{route}"
        else:
            result = f"{protocol}://{host}:{port}/{route}"
        return result

    def __str__(self) -> str:
        return self.resolve()


# =====================================================================================================================



import time

import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_datetime.m2_datetime import *
import operator


# =====================================================================================================================
NOW_TD = datetime.datetime.now()
NOW_TS = NOW_TD.timestamp()


# =====================================================================================================================
class Test__DateTime:
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (True, Exception),
            # (None, Exception),

            (datetime.datetime(1,2, 3, 11, 22, 33), datetime.datetime(1,2,3, 11, 22, 33)),

            (datetime.date(1,2,3) , datetime.date(1,2,3)),
            (datetime.time(11,22,33), datetime.time(11,22,33)),

            (NOW_TD, NOW_TD),
            (str(NOW_TD), NOW_TD),

            # float -----------
            (NOW_TS, NOW_TD),

            # str -------------
            ("2025.02.26 17.00.56", datetime.datetime(2025,2,26, 17, 0, 56)),
            ("2025.202.26 17.00.00", Exception),
            ("2025.02.26 25.00.00", Exception),     # inappropriate

            ("2025.02.26 17.00", Exception),        # incomplete

            ("2025.1.1 1.1.1", datetime.datetime(2025, 1, 1, 1, 1, 1)),
        ]
    )
    def test__init(self, source, _EXPECTED):
        func_link = lambda: DateTimeAux(source).SOURCE
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, other, cmp_func, _EXPECTED",
        argvalues=[
            (True, True, operator.eq, Exception),
            ("2025.1.1 1.1.1", True, operator.eq, Exception),
            ("2025.1.1 1.1.1", "2025.2.", operator.lt, Exception),

            ("2025.1.1 1.1.1", "2025.1.1 1.1.1", operator.eq, True),
            ("2025.1.1 1.1.1", "2025.1.1 2.1.1", operator.lt, True),

            ("2025.1.1 1.1.1", "1.2.1", operator.lt, True),
            ("2025.1.1 1.1.1", "1.0.1", operator.lt, False),

            ("2025.1.1 1.1.1", "2025.2.1", operator.lt, True),
            ("2025.1.2 1.1.1", "2025.1.1", operator.lt, False),

            (NOW_TD, NOW_TD, operator.eq, True),
            (NOW_TD, NOW_TS, operator.eq, True),
            (NOW_TS, NOW_TD, operator.eq, True),
            (NOW_TS, NOW_TS, operator.eq, True),

            (NOW_TS, str(NOW_TS), operator.eq, True),
            (str(NOW_TS), NOW_TS, operator.eq, True),
            (str(NOW_TS), str(NOW_TS), operator.eq, True),
        ]
    )
    def test__cmp(self, source, other, cmp_func, _EXPECTED):
        func_link = lambda: cmp_func(DateTimeAux(source), other)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    def test__cmp_ms(self):
        victim1 = DateTimeAux()
        victim2 = DateTimeAux()
        assert victim1 == victim2
        assert victim1.DTm == victim2.DTm

        victim1 = DateTimeAux()
        time.sleep(0.1)
        victim2 = DateTimeAux()
        print(victim1.DTm)
        print(victim2.DTm)
        # assert str(victim1) == str(victim2)
        assert victim1.DTm != victim2.DTm
        assert victim1 != victim2

    def test__update_on_str(self):
        # victim = DateTimeAux(update_on_str=True, def_str_pattern="DTm")
        # for sleep in [0, 0.2]:
        #     str1 = str(victim)
        #     if sleep:
        #         time.sleep(sleep)
        #     str2 = str(victim)
        #     print(str1)
        #     print(str2)
        #     if sleep:
        #         assert str1 != str2
        #     else:
        #         assert str1 == str2   # not always work!

        victim = DateTimeAux(update_on_str=True, def_str_pattern="DTm")
        str1 = str(victim)
        time.sleep(0.1)
        str2 = str(victim)
        print(str1)
        print(str2)
        assert str1 != str2


# =====================================================================================================================



from typing import *
import pathlib
import logging
from logging.handlers import RotatingFileHandler


# =====================================================================================================================
def _example_zero():
    LEVEL = logging.DEBUG
    PATTERN = '%(asctime)s[%(levelname)s](%(name)s(%(filename)s).%(funcName)s/thread%(thread)s(line%(lineno)d))%(msg)s'
    formatter = logging.Formatter(PATTERN)

    logger = logging.getLogger()
    logger.setLevel(LEVEL)

    handler_file = logging.FileHandler(f"SerialServer_Base.log")
    handler_file.setLevel(LEVEL)
    handler_file.setFormatter(formatter)
    logger.addHandler(handler_file)

    handler_stream = logging.StreamHandler()
    handler_stream.setLevel(LEVEL)
    handler_stream.setFormatter(formatter)
    logger.addHandler(handler_stream)

    # loggers.log(10, 'ЛОГ10')   #DEBUG=ЛОГ10
    # loggers.log(11, 'ЛОГ11')   #Level 11=ЛОГ11
    logger.warning('MSG_DEFAULT')     # 2024-03-19 15:27:21,644[WARNING](root(t001.py).<module>/thread18536(line21))MSG_DEFAULT


# =====================================================================================================================
class Logger:
    """
    GOAL
    ----
    keep all mostly used meth in hidden class

    NOTE
    ----
    DONT use directly root as secondary loggers - use it only as first or it would be applyed automatically!
    """

    # SETTINGS ---------------------------------------
    LOG_NAME: None | str = None  # None=self "root"="root", if not StrNone=get ClassName!

    LOG_LEVEL: int = logging.DEBUG
    LOG_PATTERN: str = '%(asctime)s[%(levelname)s]%(name)s(%(filename)s).%(funcName)s(line%(lineno)d)/thread%(thread)s::%(msg)s'

    LOG_DIRPATH: None | pathlib.Path = None

    LOG_FILE_STARTWITH: None | str = None
    LOG_FILE_EXTENTION: None | str = None
    LOG_FILE_MAXBYTES: int = 1024 * 1024 * 10
    LOG_FILE_BACKUPCOUNT: int = 3

    LOG_ENABLE: None | bool = False     # you need to turnON if you need it!
    LOG_USE_STREAM: bool = True
    LOG_USE_FILE: bool = False

    # AUX ---------------------------------------
    _formatter: logging.Formatter
    LOGGER: logging.Logger | None = None

    _handler_stream: logging.StreamHandler | None = None
    _handler_file: logging.FileHandler | None = None

    @property
    def LOG_FILENAME(self) -> str:
        return f"{self.LOG_FILE_STARTWITH or 'loggers'}__{self.LOG_NAME or 'root'}.{self.LOG_FILE_EXTENTION or 'log'}"

    @property
    def LOG_FILEPATH(self) -> pathlib.Path:
        if self.LOG_DIRPATH is None:
            return pathlib.Path(self.LOG_FILENAME)
        else:
            return self.LOG_DIRPATH.joinpath(self.LOG_FILENAME)

    def __init__(
            self,
            log_name: None | str | Any = None,
            log_dirpath: None | str | pathlib.Path = None,
            log_level: None | int = None,
            log_use_stream: None | bool = None,
            log_use_file: None | bool = None,
            log_enable: None | bool = None,

            *args,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # PARAMS ------------------------------------------
        if log_name is not None:
            self.LOG_NAME = log_name

        if log_dirpath is not None:
            self.LOG_DIRPATH = pathlib.Path(log_dirpath)

        if log_level is not None:
            self.LOG_LEVEL = log_level

        if log_use_stream is not None:
            self.LOG_USE_STREAM = log_use_stream

        if log_use_file is not None:
            self.LOG_USE_FILE = log_use_file

        if log_enable is not None:
            self.LOG_ENABLE = log_enable

        # PREPARE ------------------------------------------
        # if log_name as istance!
        if self.LOG_NAME is None:
            class_name = self.__class__.__name__
            if class_name == "Logger":
                self.LOG_NAME = "root"
            else:
                self.LOG_NAME = class_name
        elif not isinstance(self.LOG_NAME, str):
            # if set name by passing object
            self.LOG_NAME = self.LOG_NAME.__class__.__name__

        # if not self.__class__.LOGGER:
        #     # place here MRO name??? for classmethods???
        #     # useful for methods starts after inited first instance
        #     self.__class__.LOGGER = logging.getLogger(self.LOG_NAME)

        self.LOGGER = logging.getLogger(self.LOG_NAME)

        # DISABLE ------------------------------------------
        if self.LOG_NAME == "root" and log_enable:
            pass

        elif not self.LOG_ENABLE:
            return

        if self.LOGGER.handlers:  # already created loggers from previous inition
            return

        # --------------------------------------------------
        # TODO: create not exists LOG_DIRPATH

        # INIT ---------------------------------------------
        self.LOGGER.setLevel(self.LOG_LEVEL)

        self._formatter = logging.Formatter(self.LOG_PATTERN)

        # CONNECT -------------------------------------------
        if self.LOG_USE_STREAM:
            self.__class__._handler_stream = logging.StreamHandler()
            # self.__class__._handler_stream.setLevel(self.LOG_LEVEL)
            self.__class__._handler_stream.setFormatter(self._formatter)
            self.LOGGER.addHandler(self.__class__._handler_stream)

        if self.LOG_USE_FILE:
            self._handler_file = RotatingFileHandler(self.LOG_FILEPATH, maxBytes=self.LOG_FILE_MAXBYTES, backupCount=self.LOG_FILE_BACKUPCOUNT)
            # self._handler_file.setLevel(self.LOG_LEVEL)
            self._handler_file.setFormatter(self._formatter)
            self.LOGGER.addHandler(self._handler_file)

        # INIT ROOT -----------------------------------------
        # run always after connects!
        self._log_init_root()

        # INITIAL MSG ---------------------------------------
        self.LOGGER.debug("="*100)

        if self.LOG_USE_STREAM:
            self.LOGGER.debug(f"[Logger.{self.LOG_NAME}] start STREAM")

        if self.LOG_USE_FILE:
            self.LOGGER.debug(f"[Logger.{self.LOG_NAME}] start FILE=[{self._handler_file.baseFilename}]")

    @staticmethod
    def _log_init_root(log_enable: bool = None) -> None:
        """
        DONT USE IT DIRECTLY!!!
        it would used always automated!!!
        """
        if not log_enable:
            return
        logger_root = logging.getLogger()
        if not logger_root.hasHandlers():
            Logger(log_enable=log_enable)  # DONT USE cls()!!!


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



# =====================================================================================================================
class Symbols:
    # FINAM COMMON AVAILABLE
    # see https://docs.comon.ru/general-information/schedule-and-financial-instruments/

    BRENT_UNIVERSAL: str = "Нефть Brent"

    SYMBOLS__SORTED_TRADED_VOLUME_CURRENCY: set[str] = {
        # from SYMBOLS__RUS_FINAM!!!!
        # + https://www.moex.com/ru/marketdata/?g=4#/mode=groups&group=4&collection=3&boardgroup=57&data_type=current&category=main

        # START
        'GAZP', 'SBER', 'LKOH', 'PLZL',

        # 1 MRD
        'MGNT', 'NVTK', 'YNDX', 'VTBR', 'GMKN', 'SNGS',

        # 500 MLN
        'ROSN',
        'VKCO',
        'CHMF',

        # 400 MLN
        # 'TRUR',     # БПИФ ТинькофВечныйПортфель
        'SIBN', 'NLMK', 'ALRS', 'OZON', 'TCSG',

        # 300 MLN
        'PHOR', 'SBERP',
        # 'ORUP',     # ПАО ОР ГРУПП
        # 'KMAZ',
        # 'TGLD',     # БПИФ ТинькофЗолото
        'MOEX', 'SNGSP', 'RUAL', 'SGZH', 'FEES', 'CBOM', 'MTSS',
        "POLY",     # не попал изза иностранного названия!!!

        # 200 MLN
        'TATN', 'MAGN', 'MTLR', 'FIVE', 'AFLT', 'AFKS', 'IRAO', 'ENPG', 'VRSB', 'ABRD',
        # 'SBRB'      # СберРублевыеКорпОблигации

        # 100 MLN
    }

    SYMBOLS__BLUE: set[str] = {
        "SBER", "SBERP", "VTBR",
        "LKOH", "NVTK", "GAZP", "ROSN", "TATN", "SNGS", "SNGSP",
        "CHMF", "GMKN", "ALRS",
        "MGNT", "FIVE",
        "YNDX", "MTSS",
    }
    SYMBOLS__over100mln: set[str] = {
        *SYMBOLS__BLUE,
        "MOEX",
        "POLY", "MAGN", "MTLR", "NLMK", "PLZL", "RASP", "RUAL",
        "AKRN",
        "AFKS", "IRAO",
        "AFLT",
        "PIKK",
        "HYDR",
        "PHOR",
    }
    SYMBOLS__ALL: set[str] = {
        *SYMBOLS__over100mln
    }

    SYMBOLS__RUS_FINAM: set[str] = [
        'ABRD', 'AFKS', 'AFLT', 'AGRO', 'AKMB', 'AKME', 'AKRN', 'ALRS', 'APTK', 'AQUA', 'ARSA', 'ASSB', 'AVAN',
        'BANE', 'BANEP', 'BCSB', 'BELU', 'BISVP', 'BLNG', 'BRZL', 'BSPB', 'BSPBP',
        'CBOM', 'CHGZ', 'CHKZ', 'CHMF', 'CHMK', 'CNTL', 'CNTLP',
        'DASB', 'DIOD', 'DIVD', 'DSKY', 'DVEC', 'DZRD', 'DZRDP',
        'EELT', 'ELTZ', 'ENPG', 'ENRU', 'ESGR', 'ETLN',
        'FEES', 'FESH', 'FIVE', 'FIXP', 'FLOT', 'FMUS',
        'GAZA', 'GAZAP', 'GAZC', 'GAZP', 'GAZS', 'GAZT', 'GCHE', 'GEMA', 'GLTR', 'GMKN', 'GPBM', 'GPBS', 'GRNT', 'GTRK',
        'HHRU', 'HIMCP', 'HYDR',
        'IDVP', 'IGST', 'IGSTP', 'INGR', 'IRAO', 'IRKT', 'ISKJ',
        'JNOS', 'JNOSP',
        'KAZT', 'KAZTP', 'KBSB', 'KCHE', 'KCHEP', 'KGKC', 'KGKCP', 'KLSB', 'KMAZ', 'KMEZ', 'KMTZ', 'KOGK', 'KRKN', 'KRKNP', 'KRKOP', 'KROT',
              'KROTP', 'KRSB', 'KRSBP', 'KTSB', 'KTSBP', 'KUBE', 'KUZB', 'KZOS', 'KZOSP',
        'LIFE', 'LKOH', 'LNZL', 'LNZLP', 'LPSB', 'LSNG', 'LSNGP', 'LSRG', 'LVHK', 'LENT',
        'MAGE', 'MAGEP', 'MAGN', 'MDMG', 'MERF', 'MFGS', 'MFGSP', 'MGNT', 'MGTS', 'MGTSP', 'MISB', 'MISBP', 'MOEX', 'MRKC', 'MRKK', 'MRKP', 'MRKS', 'MRKU', 'MRKV',
              'MRKY', 'MRKZ', 'MRSB', 'MSNG', 'MSRS', 'MSTT', 'MTEK', 'MTLR', 'MTLRP', 'MTSS', 'MVID',
        'NAUK', 'NFAZ', 'NKHP', 'NKNC', 'NKNCP', 'NKSH', 'NLMK', 'NMTP', 'NNSB', 'NNSBP', 'NSVZ', 'NVTK',
        'ODVA', 'OGKB', 'OKEY', 'OMZZP', 'OZON',
        'PAZA', 'PHOR', 'PIKK', 'PLZL', 'PMSB', 'PMSBP', 'PRFN', 'PRMB', 'POSI',
        'RASP', 'RBCM', 'RCMX', 'RDRB', 'RGSS', 'RKKE', 'RNFT', 'ROLO', 'ROSB', 'ROSN', 'ROST', 'RSTI', 'RSTIP', 'RTGZ', 'RTKM', 'RTKMP',
              'RTSB', 'RTSBP', 'RUAL', 'RUGR', 'RUSI', 'RZSB', 'RENI',
        'SAGO', 'SAGOP', 'SARE', 'SAREP', 'SBER', 'SBERP', 'SBGB', 'SBMX', 'SBRB', 'SBRI', 'SELG', 'SFIN', 'SIBN', 'SLEN', 'SMLT', 'SNGS', 'SNGSP', 'STSB', 'STSBP',
              'SUGB', 'SVAV', 'SVET',
        'TASB', 'TASBP', 'TATN', 'TATNP', 'TCSG', 'TEUR', 'TGKA', 'TGKB', 'TGKBP', 'TGKN',
              'TGLD', 'TMOS', 'TNSE', 'TORS', 'TORSP', 'TRMK', 'TRNFP', 'TRUR', 'TSPX', 'TTLK', 'TUSD', 'TUZA',
        'UCSS', 'UKUZ', 'UNAC', 'UNKL', 'UPRO', 'URKZ', 'USBN', 'UTAR', 'UWGN',
        'VGSB', 'VGSBP', 'VJGZ', 'VJGZP', 'VLHZ', 'VRSB', 'VRSBP', 'VSMO', 'VSYD', 'VSYDP', 'VTBR',
        'WTCM', 'WTCMP', 'WUSH',
        'YAKG', 'YKEN', 'YKENP', 'YRSB', 'YRSBP',
        'ZILL', 'ZVEZ',

        'SGZH', 'GEMC', 'FMRU', 'VKCO', 'OBLG', 'LQDT', 'GOLD', 'EQMX', 'SPBE',
        'RCUS', 'SBGD', 'ORUP', 'SFTL']   # count = 275


# =====================================================================================================================



from base_aux.cli.m1_cli_user import *
from base_aux.versions.m2_version import *
from base_aux.aux_datetime.m2_datetime import *

# from PROJECT import PROJECT


# =====================================================================================================================
# VERSION = (0, 0, 1)   # keep russian lang by using utf-8
# VERSION = (0, 0, 2)   # reuse utf8+ del all capitalizing()
# VERSION = (0, 0, 3)   # detach dependence from importing PRJ
# VERSION = (0, 0, 4)   # use LINE_CODE_QUATATION for examples
# VERSION = (0, 0, 5)   # add BADGES block
# VERSION = (0, 0, 6)   # [BADGES] improve
# VERSION = (0, 0, 7)   # [BADGES] separate TestLinWin
# VERSION = (0, 0, 8)   # examples string add docstrings
# VERSION = (0, 0, 9)   # add gen requirements_release_freezed
# VERSION = (0, 0, 10)  # requirements_release_freezed extend timeout (at home i need 3-4sec!)
# VERSION = (0, 0, 11)  # collect all modules into one pkg!
# VERSION = (0, 0, 12)  # add bundled licenses
# VERSION = (0, 0, 13)  # bundled licenses - fix lines
VERSION = (0, 0, 14)     # move PRJBase into share +use Version


# =====================================================================================================================
class PROJECT_BASE:
    NAME_IMPORT: str
    VERSION: Version

    # AUTHOR ------------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"
    AUTHOR_NICKNAME_GITHUB: str = "centroid457"

    # AUX ----------------------------------------------------
    CLASSIFIERS_TOPICS_ADD: list[str] = [
        # "Topic :: Communications",
        # "Topic :: Communications :: Email",
    ]

    # FINALIZE -----------------------------------------------
    @classmethod
    @property
    def NAME_INSTALL(cls) -> str:
        return cls.NAME_IMPORT.replace("_", "-")


# =====================================================================================================================
class ReleaseFileBase:
    # ------------------------------------------------
    FILE_NAME: str = "FILE.md"
    PROJECT: type[PROJECT_BASE] = None

    # ------------------------------------------------
    LINE_SEPARATOR_MAIN: str = "*" * 80
    LINE_SEPARATOR_PART: str = "-" * 30

    def __init__(self, project: type[PROJECT_BASE]):
        self.PROJECT = project

    @property
    def filepath(self) -> pathlib.Path:
        return pathlib.Path(self.FILE_NAME)

    # FILE WRITE ======================================================================================================
    def _file_clear(self) -> None:
        self.filepath.write_text("")

    def _file_append_lines(self, lines: Optional[Union[str, list[str]]] = None) -> None:
        # LINES ---------------------------------
        if not lines:
            lines = ""
        if isinstance(lines, str):
            lines = [lines, ]

        # WRITE ---------------------------------
        with self.filepath.open("a", encoding="utf8") as fo_append:
            for lines in lines:
                fo_append.write(f"{lines}\n")

    # LINES ===========================================================================================================
    def _lines_create__group(self, lines: list[str], title: Optional[str] = None, nums: bool = True) -> list[str]:
        group: list[str] = []

        if title:
            group.append(title.upper())

        for num, line in enumerate(lines, start=1):
            if nums:
                bullet = f"{num}. "
            else:
                bullet = "- "

            if isinstance(line, list):
                group.append(f"{bullet}{line[0]}:  ")
                for block in line[1:]:
                    group.append(f"\t- {block}  ")
            else:
                group.append(f"{bullet}{line}  ")
        return group

    def generate(self) -> None:
        pass


# =====================================================================================================================
class ReleaseReadme(ReleaseFileBase):
    # ------------------------------------------------
    FILE_NAME: str = "README.md"

    # ------------------------------------------------
    DIRNAME_EXAMPLES: str = "EXAMPLES"
    dirpath_examples: pathlib.Path = pathlib.Path(DIRNAME_EXAMPLES)

    SEPARATOR_PATTERN = r'(\**\n+)*## USAGE EXAMPLES'
    LINE_CODE_QUATATION: str = "```"

    # GENERATE ========================================================================================================
    def generate(self) -> None:
        self._file_clear()
        self.append1_badges()
        self.append2_main()
        self.append3_examples()

    def append1_badges(self) -> None:
        lines = [
            # VER -------------
            f"![Ver/TestedPython](https://img.shields.io/pypi/pyversions/{self.PROJECT.NAME_IMPORT})",
            f"![Ver/Os](https://img.shields.io/badge/os_development-Windows-blue)  ",

            # -----------------
            f"![repo/Created](https://img.shields.io/github/created-at/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT})",
            f"![Commit/Last](https://img.shields.io/github/last-commit/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT})",
            # f"![Tests/GitHubWorkflowStatus](https://img.shields.io/github/actions/workflow/status/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT}/tests.yml)",
            f"![Tests/GitHubWorkflowStatus](https://github.com/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT}/actions/workflows/test_linux.yml/badge.svg)",
            f"![Tests/GitHubWorkflowStatus](https://github.com/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT}/actions/workflows/test_windows.yml/badge.svg)  ",

            # -----------------
            f"![repo/Size](https://img.shields.io/github/repo-size/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT})",
            *[
                f"![Commit/Count/{period}](https://img.shields.io/github/commit-activity/{period}/{self.PROJECT.AUTHOR_NICKNAME_GITHUB}/{self.PROJECT.NAME_IMPORT})"
                for period in ["t", "y", "m", ]
            ],
            f"",

        ]
        self._file_append_lines(lines)

    def append2_main(self) -> None:
        # FEATURES ----------------------------------------------------
        features = [
            f"",
            f"",
            f"## Features",
        ]
        for num, feature in enumerate(self.PROJECT.FEATURES, start=1):
            if isinstance(feature, list):
                features.append(f"{num}. {feature[0]}:  ")
                for block in feature[1:]:
                    features.append(f"\t- {block}  ")
            else:
                features.append(f"{num}. {feature}  ")

        # SUMMARY ----------------------------------------------------
        lines = [
            f"# {self.PROJECT.NAME_IMPORT} (current v{self.PROJECT.VERSION}/![Ver/Pypi Latest](https://img.shields.io/pypi/v/{self.PROJECT.NAME_IMPORT}?label=pypi%20latest))",

            f"",
            f"## DESCRIPTION_SHORT",
            f"{self.PROJECT.DESCRIPTION_SHORT.strip()}",

            f"",
            f"## DESCRIPTION_LONG",
            f"{self.PROJECT.DESCRIPTION_LONG.strip()}",

            *features,

            f"",
            f"",
            self.LINE_SEPARATOR_MAIN,
            f"## License",
            f"See the [LICENSE](LICENSE) file for license rights and limitations (MIT).  ",
            f"See the [LICENSE_bundled.md](LICENSE_bundled.md) file for parent licenses.  ",

            f"",
            f"",
            f"## Release history",
            f"See the [HISTORY.md](HISTORY.md) file for release history.  ",

            # f"",
            # f"",
            # f"## Python Installation",
            # f"```commandline",
            # f"pip install {self.PROJECT.NAME_INSTALL}",
            # f"```",

            f"",
            f"",
            f"## Requirements Installation",
            f"NOTE: all requirements for module will install automatically or install manually",
            f"```commandline",
            f"pip install -r requirements.txt",
            f"```",

            f"",
            f"",
            f"## Module Installation",
            f"```commandline",
            f"pip install {self.PROJECT.NAME_INSTALL}",
            f"```",

            f"",
            f"",
            f"## Module Import",
            f"```python",
            f"import {self.PROJECT.NAME_IMPORT}",
            f"```",
        ]
        self._file_append_lines(lines)

    def append3_examples(self) -> None:
        """
        NOTE: don't skip none-python files! it could be as part of examples! just name it in appropriate way!
        """
        LINES_EXAMPLES_START: list[str] = [
            f"",
            f"",
            self.LINE_SEPARATOR_MAIN,
            f"## USAGE EXAMPLES",
            f"See tests, sourcecode and docstrings for other examples.  ",
        ]
        self._file_append_lines(LINES_EXAMPLES_START)
        self._file_append_lines()

        files = []
        if self.dirpath_examples.exists():
            files = [item for item in self.dirpath_examples.iterdir() if item.is_file()]

        for index, file in enumerate(files, start=1):
            LINES = [
                self.LINE_SEPARATOR_PART,
                f"### {index}. {file.name}",
                self.LINE_CODE_QUATATION + ('python' if file.name.endswith(".py") else ''),
                file.read_text().strip(),
                self.LINE_CODE_QUATATION,
                f"",
            ]
            self._file_append_lines(LINES)

        self._file_append_lines(self.LINE_SEPARATOR_MAIN)


# =====================================================================================================================
class ReleaseHistory(ReleaseFileBase):
    # ------------------------------------------------
    FILE_NAME: str = "HISTORY.md"

    # ------------------------------------------------
    PATTERN_SEPARATOR_NEWS = r'#+ NEWS\s*'
    PATTERN_NEWS = r'#+ NEWS\s*(.+)\s*\*{10,}\s*'
    # PATTERN_NEWS = r'\n((?:\d+\.?){3} \((?:\d{2,4}[/:\s]?){6}\).*)\s*\*{10,}'

    LAST_NEWS: str = ""

    # PREPARE =========================================================================================================
    def load__last_news(self) -> None:
        string = self.filepath.read_text()

        # # VAR 1 --------------------------------
        # match = re.search(self.PATTERN_NEWS, string)
        # if match:
        #     self.LAST_NEWS = match[1]
        # # VAR 1 --------------------------------

        # VAR 2 --------------------------------
        splits = re.split(self.PATTERN_SEPARATOR_NEWS, string)
        if len(splits) > 1:
            self.LAST_NEWS = splits[-1]

            splits = re.split(r'\s*\*{10,}\s*', self.LAST_NEWS)
            self.LAST_NEWS = splits[0]

            splits = re.split(r'\s*0\.0\.0 \(', self.LAST_NEWS)
            self.LAST_NEWS = splits[0]

        else:
            self.LAST_NEWS = splits[0]

        # print(f"{string=}")
        # print(f"{self.LAST_NEWS=}")

    def check__new_release__is_correct(self) -> bool:
        # ----------------------------
        if self.LAST_NEWS.startswith(f"{self.PROJECT.VERSION}("):
            msg = f"exists_version {self.PROJECT.VERSION=}"
            Warn(msg)
            return False

        # ----------------------------
        for news_item in self.PROJECT.NEWS:
            if isinstance(news_item, list):
                news_item = news_item[0]

            if re.search(r'- ' + str(news_item) + r'\s*\n', self.LAST_NEWS):
                msg = f"exists_news {news_item=}"
                Warn(msg)
                return False

        # ----------------------------
        return True

    # WORK ============================================================================================================
    def lines_create__news(self) -> list[str]:
        group: list[str] = [
            f"## NEWS",
            "",
            f"{self.PROJECT.VERSION}({DateTimeAux()})",
            self.LINE_SEPARATOR_PART,
        ]
        news_new = self._lines_create__group(self.PROJECT.NEWS, nums=False)
        group.extend(news_new)
        return group

    def generate(self) -> None:
        # PREPARE --------------------------------------
        self.load__last_news()
        if not self.check__new_release__is_correct():
            msg = f"Incorrect new data (INCREASE VERSION or CHANGE NEWS)"
            raise Exc__WrongUsage_YouForgotSmth(msg)

        # WRITE ----------------------------------------
        self._file_clear()
        self.append_main()

    def append_main(self):
        lines = [
            f"# RELEASE HISTORY",
            f"",
            self.LINE_SEPARATOR_MAIN,
            *self._lines_create__group(self.PROJECT.TODO, "## TODO"),
            f"",
            self.LINE_SEPARATOR_MAIN,
            *self._lines_create__group(self.PROJECT.FIXME, "## FIXME"),
            f"",
            self.LINE_SEPARATOR_MAIN,
            *self.lines_create__news(),
            f"",
            self.LAST_NEWS,
            f"",
            self.LINE_SEPARATOR_MAIN,
        ]
        self._file_append_lines(lines)


# =====================================================================================================================
def release_files__update(project: type['PROJECT']) -> None | NoReturn:
    if len(project.VERSION) > 3:
        msg = f"Dont use more then 3 blocks in {project.VERSION=} pypi will never accept it!"
        Warn(msg)

    pass
    CliUser().send("python -m pip freeze > requirements_release_freezed.txt", timeout=10)
    ReleaseReadme(project).generate()
    ReleaseHistory(project).generate()


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================
